# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import time

from behave import *
import numpy.testing
import test

from cicada.communicator.socket import Failed, NotRunning, Revoked, SocketCommunicator
from cicada.communicator.socket.connect import EncryptionFailed, Timeout, TokenMismatch, direct, listen, rendezvous


@given(u'{} players')
def step_impl(context, players):
	context.players = eval(players)


@when(u'the players enter a barrier at different times')
def step_impl(context):
    def operation(communicator):
        time.sleep(communicator.rank * 0.1)
        enter = time.time()
        communicator.barrier()
        exit = time.time()

        return enter, exit

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, identities=context.identities, trusted=context.trusted)


@then(u'the players should exit the barrier at roughly the same time')
def step_impl(context):
    enter_delta, exit_delta = numpy.max(context.results, axis=0) - numpy.min(context.results, axis=0)
    logging.info(f"Enter delta: {enter_delta} exit delta: {exit_delta}.")
    # Every process should exit the barrier at around the same time.
    numpy.testing.assert_almost_equal(exit_delta, 0, decimal=2)


@then(u'the group result should equal {}')
def step_impl(context, result):
    result = eval(result)
    group = context.results

    test.assert_equal(result, group)


@then(u'the group should return {}')
def step_impl(context, result):
    result = numpy.array(eval(result))
    group = numpy.array(context.results)

    if issubclass(result.dtype.type, numpy.number) and issubclass(group.dtype.type, numpy.number):
        numpy.testing.assert_almost_equal(result, group, decimal=4)
    else:
        numpy.testing.assert_array_equal(result, group)


@when(u'the players allgather {values}')
def step_impl(context, values):
    values = eval(values)

    def operation(communicator, values):
        return communicator.allgather(value=values[communicator.rank])

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(values,), identities=context.identities, trusted=context.trusted)


@when(u'player {src} broadcasts {value} after the communicator has been freed')
def step_impl(context, src, value):
    src = eval(src)
    value = numpy.array(eval(value))

    def operation(communicator, src, value):
        communicator.free()
        return communicator.broadcast(src=src, value=value)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, value), identities=context.identities, trusted=context.trusted)


@when(u'player {} broadcasts {}')
def step_impl(context, src, value):
    src = eval(src)
    value = numpy.array(eval(value))

    def operation(communicator, src, value):
        return communicator.broadcast(src=src, value=value)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, value), identities=context.identities, trusted=context.trusted)


@when(u'player {dst} gathers {values} from {src}')
def step_impl(context, src, values, dst):
    src = numpy.array(eval(src))
    values = eval(values)
    dst = eval(dst)

    def operation(communicator):
        return communicator.gatherv(src=src, value=values[communicator.rank], dst=dst)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, identities=context.identities, trusted=context.trusted)


@when(u'player {dst} gathers {values}')
def step_impl(context, dst, values):
    dst = eval(dst)
    values = [value for value in numpy.array(eval(values))]

    def operation(communicator, values, dst):
        return communicator.gather(value=values[communicator.rank], dst=dst)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(values, dst), identities=context.identities, trusted=context.trusted)


@when(u'player {src} scatters messages to the other players {count} times')
def step_impl(context, src, count):
    src = eval(src)
    count = eval(count)

    def operation(communicator, src, count):
        others = set(range(communicator.world_size)) - set([src])
        for i in range(count):
            result = communicator.scatterv(src=src, values=others, dst=others)
        communicator.free()
        return communicator.stats

    context.stats = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, count), identities=context.identities, trusted=context.trusted)


@when(u'player {} scatters {} to {}')
def step_impl(context, src, values, dst):
    src = eval(src)
    values = [numpy.array(value) for value in eval(values)]
    dst = eval(dst)

    def operation(communicator, src, values, dst):
        if communicator.rank != src:
            values = None
        return communicator.scatterv(src=src, values=values, dst=dst)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, values, dst), identities=context.identities, trusted=context.trusted)


@when(u'player {} scatters {}')
def step_impl(context, src, values):
    src = eval(src)
    values = [value for value in numpy.array(eval(values))]

    def operation(communicator, src, values):
        if communicator.rank != src:
            values = None
        return communicator.scatter(src=src, values=values)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, values), identities=context.identities, trusted=context.trusted)


@then(u'player {} can send {} to player {}')
def step_impl(context, src, value, dst):
    src = eval(src)
    value = numpy.array(eval(value))
    dst = eval(dst)

    def operation(communicator, src, value, dst):
        if communicator.rank == src:
            communicator.send(value=value, dst=dst)
        if communicator.rank == dst:
            result = communicator.recv(src=src)
            numpy.testing.assert_almost_equal(value, result, decimal=4)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, value, dst), identities=context.identities, trusted=context.trusted)


@then(u'player {src} should have sent exactly {sent} messages')
def step_impl(context, src, sent):
    src = eval(src)
    sent = eval(sent)
    test.assert_equal(context.stats[src]["total"]["sent"]["messages"], sent)


@then(u'every player other than {src} should receive exactly {received} messages')
def step_impl(context, src, received):
    src = eval(src)
    received = eval(received)

    stats = [player for index, player in enumerate(context.stats) if index != src]
    for player, expected in zip(stats, received):
        numpy.testing.assert_equal(player["total"]["received"]["messages"], expected)


@then(u'it should be possible to start and stop a communicator {count} times')
def step_impl(context, count):
    count = eval(count)

    def operation(communicator):
        pass

    for i in range(count):
        SocketCommunicator.run(world_size=context.players, fn=operation, identities=context.identities, trusted=context.trusted)


@when(u'the players split the communicator with names {names}')
def step_impl(context, names):
    names = eval(names)

    def operation(communicator, groups, identities, trusted):
        comm = communicator.split(name=groups[communicator.rank], identity=identities[communicator.rank], trusted=trusted)
        if comm is not None:
            return {"name": comm.name, "world_size": comm.world_size}
        else:
            return {}

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(names, context.identities, context.trusted), family=context.family, identities=context.identities, trusted=context.trusted)


@then(u'the new communicator names should match {names}')
def step_impl(context, names):
    names = eval(names)
    test.assert_equal([result.get("name") for result in context.results], list(names))


@then(u'the new communicator world sizes should match {world_sizes}')
def step_impl(context, world_sizes):
    world_sizes = eval(world_sizes)
    test.assert_equal([result.get("world_size") for result in context.results], list(world_sizes))


@when(u'players {group} create a new communicator with world size {world_size} and name {name} and token {token}')
def step_impl(context, group, world_size, name, token):
    group = eval(group)
    world_size = eval(world_size)
    name = eval(name)
    token = eval(token)

    def operation(communicator, group, world_size, name, token):
        if communicator.rank in group:
            rank = group.index(communicator.rank)
            address = "tcp://127.0.0.1:25000" if rank == 0 else "tcp://127.0.0.1"
            listen_socket = listen(address=address, rank=communicator.rank, name=name)
            sockets=rendezvous(listen_socket=listen_socket, root_address="tcp://127.0.0.1:25000", world_size=world_size, rank=rank, name=name, token=token)
            comm = SocketCommunicator(sockets=sockets, name=name)
            return {"name": comm.name, "world_size": comm.world_size}
        return {}

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(group, world_size, name, token), identities=context.identities, trusted=context.trusted)


@when(u'players {group} create a new communicator with world size {world_size} and name {name} and tokens {tokens}')
def step_impl(context, group, world_size, name, tokens):
    group = eval(group)
    world_size = eval(world_size)
    name = eval(name)
    tokens = eval(tokens)

    def operation(communicator, group, world_size, name):
        if communicator.rank in group:
            rank = group.index(communicator.rank)
            address = "tcp://127.0.0.1:25000" if rank == 0 else "tcp://127.0.0.1"
            listen_socket = listen(address=address, rank=rank, name=name)
            sockets=rendezvous(listen_socket=listen_socket, root_address="tcp://127.0.0.1:25000", world_size=world_size, rank=rank, name=name, token=tokens[rank])
            comm = SocketCommunicator(sockets=sockets, name=name)
            return {"name": comm.name, "world_size": comm.world_size}
        return {}

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(group, world_size, name), identities=context.identities, trusted=context.trusted)


@when(u'players {group} create a new communicator with name {name} and direct addresses {addresses}')
def step_impl(context, group, name, addresses):
    group = eval(group)
    name = eval(name)
    addresses = eval(addresses)

    def operation(communicator, group, name, addresses):
        if communicator.rank in group:
            listen_socket = listen(address=addresses[communicator.rank], name=name, rank=communicator.rank)
            sockets = direct(listen_socket=listen_socket, addresses=addresses, rank=communicator.rank)
            comm = SocketCommunicator(sockets=sockets, name=name)
            return {"name": comm.name, "world_size": comm.world_size}
        return {}

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(group, name, addresses), identities=context.identities, trusted=context.trusted)


@then(u'players {players} should timeout')
def step_impl(context, players):
    players = eval(players)
    for player in players:
        result = context.results[player]
        test.assert_is_instance(result, Failed)
        test.assert_is_instance(result.exception, Timeout)


@then(u'players {players} should fail with TokenMismatch errors')
def step_impl(context, players):
    players = eval(players)
    for player in players:
        result = context.results[player]
        test.assert_is_instance(result, Failed)
        test.assert_is_instance(result.exception, TokenMismatch)


@then(u'shrinking the communicator under normal conditions will return the same players in the same rank order')
def step_impl(context):
    def operation(communicator, identities, trusted):
        comm, newranks = communicator.shrink(name="split", identity=identities[communicator.rank], trusted=trusted)
        return(newranks)

    results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(context.identities, context.trusted), identities=context.identities, trusted=context.trusted)
    for result in results:
        test.assert_equal(result, list(range(context.players)))


@when(u'players {group} shrink the communicator with name {name}')
def step_impl(context, group, name):
    group = eval(group)
    name = eval(name)
    def operation(communicator, group, name, identities, trusted):
        if communicator.rank in group:
            comm, newranks = communicator.shrink(name=name, identity=identities[communicator.rank], trusted=trusted)
            return {"name": comm.name, "world_size": comm.world_size}
        return {}

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(group, name, context.identities, context.trusted), family=context.family, identities=context.identities, trusted=context.trusted)


@when(u'player {player} revokes the communicator')
def step_impl(context, player):
    player = eval(player)
    def operation(communicator, player):
        while True:
            try:
                value = communicator.broadcast(src=0, value="foo")
                if communicator.rank == player:
                    communicator.revoke()
            except Revoked as e:
                raise e
            except Exception as e:
                logging.error(f"Player {communicator.rank} exception: {e}")

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(player,))


@then(u'players {players} should fail with NotRunning errors')
def step_impl(context, players):
    players = eval(players)
    for player in players:
        result = context.results[player]
        test.assert_is_instance(result, Failed)
        test.assert_is_instance(result.exception, NotRunning)


@then(u'players {players} should fail with Revoked errors')
def step_impl(context, players):
    players = eval(players)
    for player in players:
        result = context.results[player]
        test.assert_is_instance(result, Failed)
        test.assert_is_instance(result.exception, Revoked)


@when(u'the communicator timeout is permanently changed to {timeout}')
def step_impl(context, timeout):
    timeout = eval(timeout)

    def operation(communicator, timeout):
        timeout1 = communicator.timeout
        communicator.timeout = timeout
        timeout2 = communicator.timeout

        return [timeout1, timeout2]

    context.timeouts = numpy.array(SocketCommunicator.run(world_size=context.players, fn=operation, args=(timeout,), identities=context.identities, trusted=context.trusted))


@when(u'the communicator timeout is temporarily changed to {timeout}')
def step_impl(context, timeout):
    timeout = eval(timeout)

    def operation(communicator, timeout):
        timeout1 = communicator.timeout
        with communicator.override(timeout=timeout):
            timeout2 = communicator.timeout
        timeout3 = communicator.timeout

        return [timeout1, timeout2, timeout3]

    context.timeouts = numpy.array(SocketCommunicator.run(world_size=context.players, fn=operation, args=(timeout,), identities=context.identities, trusted=context.trusted))


@then(u'the initial communicator timeouts should match {timeouts}')
def step_impl(context, timeouts):
    timeouts = eval(timeouts)
    numpy.testing.assert_array_equal(timeouts, context.timeouts[:, 0])


@then(u'the temporary communicator timeouts should match {timeouts}')
def step_impl(context, timeouts):
    timeouts = eval(timeouts)
    numpy.testing.assert_array_equal(timeouts, context.timeouts[:, 1])


@then(u'the final communicator timeouts should match {timeouts}')
def step_impl(context, timeouts):
    timeouts = eval(timeouts)
    numpy.testing.assert_array_equal(timeouts, context.timeouts[:, -1])


@when(u'the players create a new communicator with connect.')
def step_impl(context):
    def operation(communicator):
        comm = communicator.connect(
            world_size=communicator.world_size,
            rank=communicator.rank,
            address="tcp://127.0.0.1:25252" if communicator.rank == 0 else "tcp://127.0.0.1",
            root_address="tcp://127.0.0.1:25252",
            )

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, identities=context.identities, trusted=context.trusted)


@when(u'the players create a new communicator with connect using environment variables.')
def step_impl(context):
    def operation(communicator):
        os.environ["CICADA_WORLD_SIZE"] = str(communicator.world_size)
        os.environ["CICADA_RANK"] = str(communicator.rank)
        os.environ["CICADA_ADDRESS"] = "tcp://127.0.0.1:25252" if communicator.rank == 0 else "tcp://127.0.0.1"
        os.environ["CICADA_ROOT_ADDRESS"] = "tcp://127.0.0.1:25252"

        with communicator.connect() as comm:
            pass

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, identities=context.identities, trusted=context.trusted)


@when(u'the players create a new communicator with connect using environment variables and tls.')
def step_impl(context):
    def operation(communicator, identities, trusted):
        os.environ["CICADA_WORLD_SIZE"] = str(communicator.world_size)
        os.environ["CICADA_RANK"] = str(communicator.rank)
        os.environ["CICADA_ADDRESS"] = "tcp://127.0.0.1:25252" if communicator.rank == 0 else "tcp://127.0.0.1"
        os.environ["CICADA_ROOT_ADDRESS"] = "tcp://127.0.0.1:25252"
        os.environ["CICADA_IDENTITY"] = identities[communicator.rank]
        os.environ["CICADA_TRUSTED"] = ",".join(trusted)

        with communicator.connect() as comm:
            pass

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(context.identities, context.trusted), identities=context.identities, trusted=context.trusted)


@when(u'player {src} asynchronously sends {value} to player {dst} who waits')
def step_impl(context, src, value, dst):
    src = eval(src)
    value = eval(value)
    dst = eval(dst)

    def operation(communicator, src, value, dst):
        if communicator.rank == src:
            time.sleep(0.5)
            result = communicator.isend(value=value, dst=dst, tag=13)
            assert(result.is_completed)
            result.wait()
        elif communicator.rank == dst:
            result = communicator.irecv(src=src, tag=13)
            result.wait()
            return result.value

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, value, dst), identities=context.identities, trusted=context.trusted)


@when(u'player {src} asynchronously sends {value} to player {dst}')
def step_impl(context, src, value, dst):
    src = eval(src)
    value = eval(value)
    dst = eval(dst)

    def operation(communicator, src, value, dst):
        if communicator.rank == src:
            time.sleep(0.5)
            result = communicator.isend(value=value, dst=dst, tag=13)
            assert(result.is_completed)
            result.wait()
        elif communicator.rank == dst:
            result = communicator.irecv(src=src, tag=13)
            while True:
                if result.is_completed:
                    return result.value
                time.sleep(0.1)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(src, value, dst), identities=context.identities, trusted=context.trusted)


@when(u"the players create a new communicator with connect, but player {a} doesn't trust player {b}")
def step_impl(context, a, b):
    a = eval(a)
    b = eval(b)

    def operation(communicator, identities, trusted, a, b):
        if a == communicator.rank:
            del trusted[b]

        comm = communicator.connect(
            world_size=communicator.world_size,
            rank=communicator.rank,
            address="tcp://127.0.0.1:25252" if communicator.rank == 0 else "tcp://127.0.0.1",
            root_address="tcp://127.0.0.1:25252",
            identity=identities[communicator.rank],
            trusted=trusted,
            )

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(context.identities, context.trusted, a, b), identities=context.identities, trusted=context.trusted)


@then(u'the group should raise exceptions {exceptions}')
def step_impl(context, exceptions):
    exceptions = eval(exceptions)

    test.assert_equal(len(context.results), len(exceptions))
    for lhs, rhs in zip(context.results, exceptions):
        if rhs is None:
            test.assert_equal(lhs, rhs)
        else:
            test.assert_is_instance(lhs, Failed)
            test.assert_is_instance(lhs.exception, rhs)


@given(u'{family} addressing')
def step_impl(context, family):
    family = eval(family)
    context.family = family


@when(u'player {slacker} fails and the others revoke and shrink the communicator, the new communicator should include the remaining players')
def step_impl(context, slacker):
    slacker = eval(slacker)

    def operation(communicator, slacker):
        pass

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(slacker,), identities=context.identities, trusted=context.trusted)


@when(u'every player raises {exception}')
def step_impl(context, exception):
    exception = eval(exception)

    def operation(communicator, exception):
        raise exception

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(exception,), identities=None, trusted=None, startup_timeout=context.startup_timeout)


@then(u'SocketCommunicator.run should catch {exception} from every player')
def step_impl(context, exception):
    exception = eval(exception)

    lhs = [repr(exception)] * context.players
    rhs = [repr(result.exception) for result in context.results]
    test.assert_equal(lhs, rhs)


@given(u'a startup timeout of {timeout} seconds')
def step_impl(context, timeout):
    timeout = eval(timeout)
    context.startup_timeout = timeout




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "Cicada"
copyright = "2021, National Technology & Engineering Solutions of Sandia, LLC (NTESS)"
author = "Timothy M. Shead"

# The full version, including alpha/beta/rc tags
import cicada
release = cicada.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.

import sphinx_rtd_theme

extensions = [
#    "IPython.sphinxext.ipython_console_highlighting",
    "nbsphinx",
    "sphinxarg.ext",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_gallery.load_style",
    "sphinx_rtd_theme",
]

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "python": ("https://docs.python.org/3", None),
    }

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

nitpicky = True

# -- nbsphinx options --------------------------------------------------------

nbsphinx_execute = "never"

nbsphinx_thumbnails = {
    "user-guide/absolute": "_static/absolute.png",
    "user-guide/active": "_static/logo.png",
    "user-guide/bit_decomposition": "_static/logo.png",
    "user-guide/division": "_static/division.png",
    "user-guide/equality": "_static/equality.png",
    "user-guide/faults": "_static/logo.png",
    "user-guide/fields-and-semantics": "_static/logo.png",
    "user-guide/floor": "_static/floor.png",
    "user-guide/interactive-programs": "_static/logo.png",
    "user-guide/less-than": "_static/less-than.png",
    "user-guide/less-than-zero": "_static/less-than-zero.png",
    "user-guide/logging": "_static/logo.png",
    "user-guide/logical-not": "_static/logical-not.png",
    "user-guide/logical-xor": "_static/logical-xor.png",
    "user-guide/modulus": "_static/modulus.png",
    "user-guide/multiple-communicators": "_static/logo.png",
    "user-guide/multiplicative-inverse": "_static/multiplicative-inverse.png",
    "user-guide/multiplication-and-truncation": "_static/multiplication.png",
    "user-guide/power": "_static/power.png",
    "user-guide/random-bits": "_static/logo.png",
    "user-guide/random-numbers": "_static/logo.png",
    "user-guide/random-seeds": "_static/logo.png",
    "user-guide/relu": "_static/relu.png",
    "user-guide/running-cicada-programs": "_static/logo.png",
    "user-guide/self-conscious-statisticians": "_static/logo.png",
    "user-guide/shamir": "_static/logo.png",
    "user-guide/sockets": "_static/logo.png",
    "user-guide/transcripts": "_static/logo.png",
    "user-guide/zigmoid": "_static/zigmoid.png",
    "user-guide/communication-patterns": "_static/communication-patterns_generic.png",
    "user-guide/timeouts": "_static/timeouts.png",
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

def warn_undocumented_members(app, what, name, obj, options, lines):
    if what not in [] and len(lines) == 0:
        print("WARNING: %s is undocumented: %s" % (what, name))
        lines.append(".. Warning:: %s '%s' undocumented" % (what, name))

def setup(app):
    app.connect("autodoc-process-docstring", warn_undocumented_members);

tls_verify = False



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import pkgutil
import sys

from behave import *

import IPython
import nbformat


root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
docs_dir = os.path.join(root_dir, "docs")
package_dir = os.path.join(root_dir, "cicada")


@given(u'all public modules')
def step_impl(context):
    def walk_modules(package, path):
        modules = []
        modules.append(package)
        for loader, name, is_package in pkgutil.iter_modules([path]):
            modules += walk_modules(package + "." + name, os.path.join(path, name))
        return modules
    context.modules = sorted(walk_modules("cicada", package_dir))


@given(u'the reference documentation')
def step_impl(context):
    context.references = []
    for directory, subdirectories, filenames in os.walk(docs_dir):
        for filename in filenames:
            if os.path.splitext(filename)[1] not in [".rst"]:
                continue
            if not filename.startswith("cicada."):
                continue

            context.references.append(os.path.join(directory, filename))
    context.references = sorted(context.references)


@then(u'every module must have a section in the reference documentation')
def step_impl(context):
    for module in context.modules:
        if os.path.join(docs_dir, module + ".rst") not in context.references:
            raise AssertionError("No matching documentation found for the %s module." % module)


@then(u'every section in the reference documentation must match a module')
def step_impl(context):
    modules = [os.path.join(docs_dir, module + ".rst") for module in context.modules]
    for reference in context.references:
        if reference not in modules:
            raise AssertionError("No matching module found for %s." % reference)


@given(u'all documentation notebooks')
def step_impl(context):
    context.notebooks = [path for path in glob.glob("docs/**/*.ipynb", recursive=True) if "_build" not in path]


@then(u'every notebook code cell should have been executed.')
def step_impl(context):
    unexecuted_notebooks = set()

    for path in context.notebooks:
        notebook = nbformat.read(path, as_version=4)
        for cell in notebook.cells:
            if cell.cell_type == "code":
                if not cell["execution_count"]:
                    unexecuted_notebooks.add(path)

    if unexecuted_notebooks:
        raise AssertionError(f"Some notebooks contain unexecuted cells: {', '.join(sorted(unexecuted_notebooks))}")


@then(u'every notebook runs without error')
def step_impl(context):
    for notebook in context.notebooks:
        context.execute_steps(f"Then notebook {notebook} runs without error")


@then(u'notebook {notebook} runs without error')
def step_impl(context, notebook):
    notebook = os.path.abspath(notebook)
    notebook_dir = os.path.dirname(notebook)
    working_dir = os.getcwd()

    sys.path.append(notebook_dir)
    os.chdir(notebook_dir)

    exception = None
    try:
        with open(notebook) as stream:
            notebook = nbformat.read(stream, as_version=4)

        shell = IPython.core.interactiveshell.InteractiveShell.instance()
        nblocals = dict()

        for cell in notebook.cells:
            if cell.cell_type == "code":
                code = shell.input_transformer_manager.transform_cell(cell.source)
                exec(code, nblocals)
    except Exception as e:
        exception = e

    os.chdir(working_dir)
    sys.path.remove(notebook_dir)

    if exception is not None:
        raise exception



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a command line interface for running Cicada performance benchmarks.
"""

import argparse
import logging
import os
import sys
import time

import numpy

import cicada
from cicada.additive import AdditiveProtocolSuite
from cicada.communicator import SocketCommunicator


class Timer(object):
    def __init__(self):
        self._start = time.time()

    def elapsed(self):
        return time.time() - self._start


def identities(arguments):
    results = []
    for rank in range(arguments.world_size):
        identity = arguments.identity.format(rank=rank, world_size=arguments.world_size)
        if os.path.exists(identity):
            results.append(identity)
    if results:
        return results


def trusted(arguments):
    results = []
    for rank in range(arguments.world_size):
        trusted = arguments.trusted.format(rank=rank, world_size=arguments.world_size)
        if os.path.exists(trusted):
            results.append(trusted)
    if results:
        return results


def print_times(case, times):
    print(f"{case} min: {numpy.min(times):.3f}s mean: {numpy.mean(times):.3f}s max: {numpy.max(times):.3f}s stddev: {numpy.std(times):.3f}s")


def nonnegative_integer(string):
    value = int(string)
    if value < 1:
        raise ValueError("A positive integer is required.")
    return value


def positive_integer(string):
    value = int(string)
    if value < 1:
        raise ValueError("A positive integer is required.")
    return value


parser = argparse.ArgumentParser(description="Cicada MPC performance tests.")
subparsers = parser.add_subparsers(title="commands (choose one)", dest="command")

# broadcast
broadcast_subparser = subparsers.add_parser("broadcast", help="Test broadcast performance on the local machine.")
broadcast_subparser.add_argument("--count", default=10000, type=positive_integer, help="Number of broadcast operations. Default: %(default)s")
broadcast_subparser.add_argument("--identity", default="player-{rank}.pem", help="Player private key and certificate in PEM format. Default: %(default)s file, if it exists.")
broadcast_subparser.add_argument("--seed", default=1234, type=positive_integer, help="Random seed. Default: %(default)s")
broadcast_subparser.add_argument("--size", default=1000, type=positive_integer, help="Broadcast message size in bytes. Default: %(default)s")
broadcast_subparser.add_argument("--src", default=0, type=nonnegative_integer, help="Broadcast src player. Default: %(default)s")
broadcast_subparser.add_argument("--trusted", default="player-{rank}.cert", help="Trusted certificates in PEM format. Default: %(default)s files, if they exist.")
broadcast_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")

# floor
floor_subparser = subparsers.add_parser("floor", help="Test floor() performance on the local machine.")
floor_subparser.add_argument("--count", default=1, type=positive_integer, help="Number of floor() operations. Default: %(default)s")
floor_subparser.add_argument("--identity", default="player-{rank}.pem", help="Player private key and certificate in PEM format. Default: %(default)s file, if it exists.")
floor_subparser.add_argument("--seed", default=1234, type=positive_integer, help="Random seed. Default: %(default)s")
floor_subparser.add_argument("--trusted", default="player-{rank}.cert", help="Trusted certificates in PEM format. Default: %(default)s files, if they exist.")
floor_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")

# gather
gather_subparser = subparsers.add_parser("gather", help="Test gather performance on the local machine.")
gather_subparser.add_argument("--count", default=10000, type=positive_integer, help="Number of gather operations. Default: %(default)s")
gather_subparser.add_argument("--dst", default=0, type=nonnegative_integer, help="Gather dst player. Default: %(default)s")
gather_subparser.add_argument("--identity", default="player-{rank}.pem", help="Player private key and certificate in PEM format. Default: %(default)s file, if it exists.")
gather_subparser.add_argument("--seed", default=1234, type=positive_integer, help="Random seed. Default: %(default)s")
gather_subparser.add_argument("--size", default=1000, type=positive_integer, help="Gather message size in bytes. Default: %(default)s")
gather_subparser.add_argument("--trusted", default="player-{rank}.cert", help="Trusted certificates in PEM format. Default: %(default)s files, if they exist.")
gather_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")

# scatterv
scatterv_subparser = subparsers.add_parser("scatterv", help="Test scatterv performance on the local machine.")
scatterv_subparser.add_argument("--count", default=10000, type=positive_integer, help="Number of scatterv operations. Default: %(default)s")
scatterv_subparser.add_argument("--src", default=0, type=nonnegative_integer, help="Scatterv src player. Default: %(default)s")
scatterv_subparser.add_argument("--identity", default="player-{rank}.pem", help="Player private key and certificate in PEM format. Default: %(default)s file, if it exists.")
scatterv_subparser.add_argument("--seed", default=1234, type=positive_integer, help="Random seed. Default: %(default)s")
scatterv_subparser.add_argument("--size", default=1000, type=positive_integer, help="Scatterv message size in bytes. Default: %(default)s")
scatterv_subparser.add_argument("--trusted", default="player-{rank}.cert", help="Trusted certificates in PEM format. Default: %(default)s files, if they exist.")
scatterv_subparser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")

# version
version_subparser = subparsers.add_parser("version", help="Print the Cicada version.")



def main():
    arguments = parser.parse_args()

    if arguments.command is None:
        parser.print_help()

    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger()
    log.name = os.path.basename(sys.argv[0])

    # broadcast
    if arguments.command == "broadcast":

        def implementation(communicator, count, seed, size, src):
            generator = numpy.random.default_rng(seed=seed)
            timer = Timer()
            for index in range(count):
                value = generator.bytes(size) if communicator.rank == src else None
                value = communicator.broadcast(src=src, value=value)
            return timer.elapsed()

        case = f"{arguments.world_size} players broadcast {arguments.count} times using {arguments.size} byte messages"
        times = SocketCommunicator.run(world_size=arguments.world_size, fn=implementation, identities=identities(arguments), trusted=trusted(arguments), kwargs=dict(count=arguments.count, seed=arguments.seed, size=arguments.size, src=arguments.src))
        print_times(case, times)

    # floor
    if arguments.command == "floor":

        def implementation(communicator, count, seed):
            protocol = AdditiveProtocolSuite(communicator=communicator)
            generator = numpy.random.default_rng(seed=seed)
            timer = Timer()
            for index in range(count):
                value = generator.uniform(low=-5, high=5, size=()) if communicator.rank == 0 else None
                value_share = protocol.share(src=0, secret=value, shape=())
                floor_share = protocol.floor(value_share)
            return timer.elapsed()

        case = f"{arguments.world_size} players compute floor() {arguments.count} times"
        times = SocketCommunicator.run(world_size=arguments.world_size, fn=implementation, identities=identities(arguments), trusted=trusted(arguments), kwargs=dict(count=arguments.count, seed=arguments.seed))
        print_times(case, times)

    # gather
    if arguments.command == "gather":

        def implementation(communicator, count, seed, size, dst):
            generator = numpy.random.default_rng(seed=seed)
            timer = Timer()
            for index in range(count):
                value = generator.bytes(size)
                value = communicator.gather(dst=dst, value=value)
            return timer.elapsed()

        case = f"{arguments.world_size} players gather {arguments.count} times using {arguments.size} byte messages"
        times = SocketCommunicator.run(world_size=arguments.world_size, fn=implementation, identities=identities(arguments), trusted=trusted(arguments), kwargs=dict(count=arguments.count, seed=arguments.seed, size=arguments.size, dst=arguments.dst))
        print_times(case, times)

    # scatterv
    if arguments.command == "scatterv":

        def implementation(communicator, count, seed, size, src, dst):
            generator = numpy.random.default_rng(seed=seed)
            timer = Timer()
            for index in range(count):
                values = [generator.bytes(size) for rank in dst] if communicator.rank == src else None
                value = communicator.scatterv(src=src, values=values, dst=dst)
            return timer.elapsed()

        case = f"{arguments.world_size} players scatterv {arguments.count} times using {arguments.size} byte messages"
        dst = [rank for rank in range(arguments.world_size) if rank != arguments.src]
        times = SocketCommunicator.run(world_size=arguments.world_size, fn=implementation, identities=identities(arguments), trusted=trusted(arguments), kwargs=dict(count=arguments.count, seed=arguments.seed, size=arguments.size, src=arguments.src, dst=dst))
        print_times(case, times)

    # version
    if arguments.command == "version":
        print(cicada.__version__)





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import shutil
import subprocess
import sys

for path in glob.glob(".coverage*"):
    if path not in [".coveragerc"]:
        os.remove(path)
if os.path.exists(".cover"):
    shutil.rmtree(".cover")
subprocess.call(["coverage", "run", "-m", "behave"] + sys.argv[1:])
subprocess.call(["coverage", "combine"])
subprocess.call(["coverage", "report"])
subprocess.call(["coverage", "html", "--directory", ".cover"])



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for creating, manipulating, and revealing additive-shared secrets."""

import inspect
import math

import numpy

from cicada.arithmetic import Field
from cicada.communicator.interface import Communicator
from cicada.encoding import FixedPoint, Identity, Boolean
from cicada.przs import PRZSProtocol


class AdditiveArrayShare(object):
    """Stores the local share of a secret shared array for :class:`AdditiveProtocolSuite`.

    Instances of :class:`AdditiveArrayShare` should only be created
    using :class:`AdditiveProtocolSuite`.
    """
    def __init__(self, storage):
        self.storage = storage


    def __repr__(self):
        return f"cicada.additive.AdditiveArrayShare(storage={self._storage})" # pragma: no cover


    def __getitem__(self, index):
        return AdditiveArrayShare(numpy.array(self._storage[index], dtype=self._storage.dtype)) # pragma: no cover


    @property
    def storage(self):
        """Private storage for the local share of a secret shared array.
        Access is provided only for serialization and communication -
        callers must use :class:`AdditiveProtocolSuite` to manipulate secret
        shares.
        """
        return self._storage


    @storage.setter
    def storage(self, storage):
        if not isinstance(storage, numpy.ndarray):
            raise ValueError(f"Expected numpy.ndarray, got {type(storage)}.") # pragma: no cover
        self._storage = numpy.array(storage, dtype=object)


class AdditiveProtocolSuite(object):
    """Protocol suite implementing computation with additive-shared secrets.

    Multiplication is implemented using a generalization of "Protocols for
    secure remote database access with approximate matching" by Du and Atallah,
    which provides semi-honest security and does not require Beaver triples or
    other offline computation.

    Comparisons are based on "Multiparty computation for interval, equality,
    and comparison without bit-decomposition protocol" by Nishide and Ohta, and
    inherit the semi-honest security model from multiplication.

    Note
    ----
    Creating the protocol is a collective operation that *must*
    be called by all players that are members of `communicator`.

    Parameters
    ----------
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that this protocol will use for communication.
    seed: :class:`int`, optional
        Seed used to initialize random number generators.  For privacy, this
        value should be different for each player.  By default, the seed will
        be chosen at random, and is guaranteed to be different even on forked
        processes.  If you specify `seed` yourself, the actual seed used will
        be the sum of this value and the value of `seed_offset`.
    seed_offset: :class:`int`, optional
        Value added to the value of `seed`.  This value defaults to the player's
        rank.
    order: :class:`int`, optional
        Field size for storing encoded values.  Defaults to the largest prime
        less than :math:`2^{64}`.
    encoding: :class:`object`, optional
        Default encoding to use for operations that require encoding/decoding.
        Uses an instance of :any:`FixedPoint` with 16 bits of
        floating-point precision if :any:`None`.
    """
    def __init__(self, communicator, *, seed=None, seed_offset=None, order=None, encoding=None):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover

        # Choose a random seed, if the caller hasn't chosen one already. The
        # chosen seed could be any number, but we choose a random 64-bit
        # integer here so other players cannot guess its value.  We typically
        # get here from a forked process, so we use numpy's random generator
        # because it will produce different seeds even from forked processes.
        if seed is None:
            seed = numpy.random.default_rng(seed=None).integers(low=0, high=2**63-1, endpoint=True)
        else:
            if seed_offset is None:
                seed_offset = communicator.rank
            seed += seed_offset

        if encoding is None:
            encoding = FixedPoint()

        field = Field(order=order)

        self._communicator = communicator
        self._field = field
        self._encoding = encoding
        self._przs = PRZSProtocol(communicator=communicator, field=field, seed=seed)


    def _assert_binary_compatible(self, lhs, rhs, lhslabel, rhslabel):
        self._assert_unary_compatible(lhs, lhslabel)
        self._assert_unary_compatible(rhs, rhslabel)
        if lhs.storage.shape != rhs.storage.shape:
            raise ValueError(f"{lhslabel} and {rhslabel} must be the same shape, got {lhs.storage.shape} and {rhs.storage.shape} instead.") # pragma: no cover


    def _assert_unary_compatible(self, share, label):
        if not isinstance(share, AdditiveArrayShare):
            raise ValueError(f"{label} must be an instance of AdditiveArrayShare, got {type(share)} instead.") # pragma: no cover


    def _require_encoding(self, encoding):
        if encoding is None:
            encoding = self._encoding
        return encoding


    def absolute(self, operand):
        """Elementwise absolute value of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared value to which the absolute value function should be applied.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise absolute value of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        ltz = self.less_zero(operand)
        nltz = self.logical_not(ltz)
        addinvop = self.negative(operand)
        ltz_parts = self.field_multiply(ltz, addinvop)
        nltz_parts = self.field_multiply(nltz, operand)
        return self.field_add(ltz_parts, nltz_parts)


    def add(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise sum of arrays.

        This method can be used to perform private-private, public-private,
        and private-public addition.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public addition
        isn't allowed, as it isn't privacy-preserving!

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be added.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be added.
        encoding: :class:`object`, optional
            Encodes public operands.  The protocol's :attr:`encoding`
            is used by default if :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared sum of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private addition.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            return self.field_add(lhs, rhs)

        # Private-public addition.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return self.field_add(lhs, encoding.encode(rhs, self.field))

        # Public-private addition.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, AdditiveArrayShare):
            return self.field_add(encoding.encode(lhs, self.field), rhs)

        raise NotImplementedError(f"Privacy-preserving addition not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def bit_compose(self, operand):
        """Compose an array of secret-shared bits into an array of corresponding integer field values.

        The result array will will have one fewer dimensions than the operand,
        which must contain bits in big-endian order in its last dimension.
        Note that the input must contain the field values :math:`0` and :math:`1`.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared array containing bits to be composed.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Share of the resulting field values.
        """
        self._assert_unary_compatible(operand, "operand")

        result = numpy.empty(operand.storage.shape[:-1], dtype=object)
        shift = numpy.power(2, numpy.arange(operand.storage.shape[-1], dtype=self.field.dtype)[::-1])
        shifted = self.field.multiply(operand.storage, shift)
        result = numpy.sum(shifted, axis=-1, out=result)
        result %= self.field.order
        return AdditiveArrayShare(result)


    def bit_decompose(self, operand, *, bits=None):
        """Decompose operand into shares of its bitwise representation.

        The result array will have one more dimension than the operand,
        containing the returned bits in big-endian order.  Note that the
        results will contain the field values :math:`0` and :math:`1`.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared array containing values to be decomposed.
        bits: :class:`int`, optional
            The number of rightmost bits in each value to extract.  Defaults to
            all bits (i.e. the number of bits used for storage by the protocol's
            :attr:`field`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Share of the bit decomposed secret.
        """
        self._assert_unary_compatible(operand, "operand")

        if bits is None:
            bits = self.field.bits
        list_o_bits = []
        two_inv = numpy.array(pow(2, self.field.order-2, self.field.order), dtype=self.field.dtype)
        for element in operand.storage.flat: # Iterates in "C" order.
            loopop = AdditiveArrayShare(numpy.array(element, dtype=self.field.dtype))
            elebits = []
            for i in range(bits):
                elebits.append(self._lsb(loopop))
                loopop = self.field_subtract(loopop, elebits[-1])
                loopop = AdditiveArrayShare(self.field.multiply(loopop.storage, two_inv))
            list_o_bits.append(elebits[::-1])
        return AdditiveArrayShare(numpy.array([x.storage for y in list_o_bits for x in y]).reshape(operand.storage.shape+(bits,)))


    @property
    def communicator(self):
        """The :class:`~cicada.communicator.interface.Communicator` used by this protocol."""
        return self._communicator


    def divide(self, lhs, rhs, *, encoding=None, rmask=None, mask1=None, rem1=None, mask2=None, rem2=None, mask3=None, rem3=None):
        """Privacy-preserving elementwise division of arrays.

        This method can be used to perform private-private and
        private-public division.  The result is the secret shared
        elementwise quotient of the operands.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared value to be divided.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be divided.
        encoding: :class:`object`, optional
            Encodes public operands and determines the number of bits to
            shift right from intermediate results.  The protocol's
            :attr:`encoding` is used by default if :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared quotient of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private division.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            zshare = self.share(src=0, secret=numpy.zeros_like(rhs.storage), shape=rhs.storage.shape)
            check = self.reveal(self.equal(rhs, zshare), encoding=Boolean())
            if not len(check.shape):
                if check:
                    raise ZeroDivisionError()
            else:
                if check.any():   
                    raise ZeroDivisionError()
            oops = True
            while oops:
                if rmask is None:
                    _, rmask = self.random_bitwise_secret(bits=encoding.precision, shape=rhs.storage.shape)
                else:
                    oops = False
                if not len(rhs.storage.shape):
                    if self.reveal(self.equal(rmask, zshare), encoding=Boolean()):
                        rmask = None
                else:
                    if self.reveal(self.equal(rmask, zshare), encoding=Boolean()).any():
                        rmask = None
            rhsmasked = self.field_multiply(rmask, rhs)
            if mask1 != None and rem1 != None:
                rhsmasked = self.right_shift(rhsmasked, bits=encoding.precision, trunc_mask=mask1, rem_mask=rem1)
            else:
                rhsmasked = self.right_shift(rhsmasked, bits=encoding.precision)
            revealrhsmasked = self.reveal(rhsmasked, encoding=encoding)
            if mask2 != None and rem2 != None:
                almost_there = self.right_shift(self.field_multiply(lhs, rmask), bits=encoding.precision, trunc_mask=mask2, rem_mask=rem2)
            else:
                almost_there = self.right_shift(self.field_multiply(lhs, rmask), bits=encoding.precision)
            divisor = encoding.encode(numpy.array(1 / revealrhsmasked), self.field)
            quotient = AdditiveArrayShare(self.field.multiply(almost_there.storage, divisor))
            if mask3 != None and rem3 != None:
                return self.right_shift(quotient, bits=encoding.precision, trunc_mask=mask3, rem_mask=rem3)
            else:
                return self.right_shift(quotient, bits=encoding.precision)

        # Private-public division.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            if rhs == numpy.zeros_like(rhs):
                raise ZeroDivisionError()
            divisor = encoding.encode(numpy.array(1 / rhs), self.field)
            result = self.field_multiply(lhs, divisor)
            result = self.right_shift(result, bits=encoding.precision)
            return result

        raise NotImplementedError(f"Privacy-preserving division not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def dot(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving dot product of two secret shared vectors.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared vector.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared vector.
        encoding: :class:`object`, optional
            Determines the number of bits to truncate from intermediate
            results.  The protocol's :attr:`encoding` is used by default if
            :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared dot product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        encoding = self._require_encoding(encoding)

        result = self.field_dot(lhs, rhs)
        result = self.right_shift(result, bits=encoding.precision)
        return result


    @property
    def encoding(self):
        """Default encoding to use for operations that require encoding/decoding."""
        return self._encoding


    def equal(self, lhs, rhs):
        """Elementwise probabilistic equality comparison between secret shared arrays.

        The result is the secret shared elementwise comparison `lhs` == `rhs`.  Note
        that the results will contain the field values :math:`0` and :math:`1`.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared array to be compared.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared array to be compared.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared result from comparing `lhs` == `rhs` elementwise.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        diff = self.field_subtract(lhs, rhs)
        return self.logical_not(self.field_power(diff, self.field.order - 1))


    @property
    def field(self):
        """Integer :any:`Field` used for arithmetic on and storage of secret shared values."""
        return self._field


    def field_add(self, lhs, rhs):
        """Privacy-preserving elementwise sum of arrays.

        This method can be used to perform private-private, public-private,
        and private-public addition.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public addition
        isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`add`, :meth:`field_add` only operates on field
        values, no encoding is performed on its inputs.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be added.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be added.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared sum of `lhs` and `rhs`.
        """
        # Private-private addition.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            return AdditiveArrayShare(self.field.add(lhs.storage, rhs.storage))

        # Private-public addition.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            if self.communicator.rank == 0:
                return AdditiveArrayShare(self.field.add(lhs.storage, rhs))
            return lhs

        # Public-private addition.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, AdditiveArrayShare):
            if self.communicator.rank == 0:
                return AdditiveArrayShare(self.field.add(lhs, rhs.storage))
            return rhs

        raise NotImplementedError(f"Privacy-preserving addition not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_dot(self, lhs, rhs):
        """Privacy-preserving dot product of two secret shared vectors.

        Unlike :meth:`dot`, :meth:`field_dot` only operates on field values,
        no right shift is performed on the results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared vector.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared vector.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared dot product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        result = self.field_multiply(lhs, rhs)
        result = self.sum(result)
        return result


    def field_multiply(self, lhs, rhs):
        """Privacy-preserving elementwise multiplication of arrays.

        This method can be used to perform private-private, public-private, and
        private-public multiplication.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public
        multiplication isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`multiply`, :meth:`field_multiply` only operates on field
        values, no encoding is performed on its inputs, and no right shift
        is performed on the results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared product of `lhs` and `rhs`.
        """
        # Private-private multiplication.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            # To multiply using additive shares X and Y, we need to compute the
            # following polynomial:
            #
            #    (X0 + X1 + ... Xn-1)(Y0 + Y1 + ... Yn-1)
            #
            # To do so, we carefully share the terms of the polynomial with the
            # other players while ensuring that no one player receives every share
            # of either secret.  Each player multiplies and sums the terms that
            # they have on hand, producing an additive share of the result.

            rank = self.communicator.rank
            world_size = self.communicator.world_size
            count = math.ceil((world_size - 1) / 2)
            x = lhs.storage
            y = rhs.storage
            X = [] # Storage for shares received from other players.
            Y = [] # Storage for shares received from other players.

            # Distribute terms to the other players.
            for src in self.communicator.ranks:
                # Identify which players will receive terms.
                if world_size % 2 == 0 and src >= count:
                    dst = numpy.arange(src + 1, src + 1 + count - 1) % world_size
                else:
                    dst = numpy.arange(src + 1, src + 1 + count) % world_size

                # Send terms to the other players.
                values = [x] * len(dst) if src == rank else None
                share = self.communicator.scatterv(src=src, dst=dst, values=values)
                if rank in dst:
                    X.append(share)
                values = [y] * len(dst) if src == rank else None
                share = self.communicator.scatterv(src=src, dst=dst, values=values)
                if rank in dst:
                    Y.append(share)

            # Multiply the polynomial terms that we have on-hand.
            result = x * y
            for other_x, other_y in zip(X, Y):
                result += x * other_y + other_x * y

            return AdditiveArrayShare(numpy.array(result % self.field.order, dtype=self.field.dtype))

        # Public-private multiplication.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, AdditiveArrayShare):
            return AdditiveArrayShare(self.field.multiply(lhs, rhs.storage))

        # Private-public multiplication.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return AdditiveArrayShare(self.field.multiply(lhs.storage, rhs))

        raise NotImplementedError(f"Privacy-preserving multiplication not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_power(self, lhs, rhs):
        """Privacy-preserving elementwise exponentiation.

        Raises secret shared array values to public integer values.  Unlike :meth:`power`,
        :meth:`field_power` only operates on field values, no right shift is performed on the
        results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared values which iwll be raised to a power.
        rhs: :class:`int` or integer :class:`numpy.ndarray`, required
            Public integer power(s) to which each element in `lhs` will be raised.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared result of raising `lhs` to the power(s) in `rhs`.
        """
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, (numpy.ndarray, int)):
            if isinstance(rhs, int):
                rhs = self.field.full_like(lhs.storage, rhs)

            ans = []
            for lhse, rhse in numpy.nditer([lhs.storage, rhs], flags=(["refs_ok"])):
                rhsbits = [int(x) for x in bin(int(rhse))[2:]][::-1]
                tmp = AdditiveArrayShare(lhse)
                it_ans = self.share(src = 0, secret=self.field.full_like(lhse, 1), shape=lhse.shape, encoding=Identity())
                limit = len(rhsbits)-1
                for i, bit in enumerate(rhsbits):
                    if bit:
                        it_ans = self.field_multiply(it_ans, tmp)
                    if i < limit:
                        tmp = self.field_multiply(tmp,tmp)
                ans.append(it_ans)
            return AdditiveArrayShare(numpy.array([x.storage for x in ans], dtype=self.field.dtype).reshape(lhs.storage.shape))

        raise NotImplementedError(f"Privacy-preserving exponentiation not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_subtract(self, lhs, rhs):
        """Privacy-preserving elementwise difference of arrays.

        This method can be used to perform private-private, public-private, and
        private-public subtraction.  The result is the secret shared
        elementwise difference of the operands.  Note that public-public
        subtraction isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`subtract`, :meth:`field_subtract` only operates on field
        values, no encoding is performed on its inputs.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be subtracted.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be subtracted.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared difference of `lhs` and `rhs`.
        """
        # Private-private subtraction.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            return AdditiveArrayShare(self.field.subtract(lhs.storage, rhs.storage))

        # Public-private subtraction.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, AdditiveArrayShare):
            if self.communicator.rank == 0:
                return AdditiveArrayShare(self.field.subtract(lhs, rhs.storage))
            return AdditiveArrayShare(self.field.negative(rhs.storage))

        # Private-public subtraction.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            if self.communicator.rank == 0:
                return AdditiveArrayShare(self.field.subtract(lhs.storage, rhs))
            return AdditiveArrayShare(lhs.storage)

        raise NotImplementedError(f"Privacy-preserving subtraction not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def field_uniform(self, *, shape=None, generator=None):
        """Generate private random field elements.

        This method can be used to generate a secret shared array containing
        random field elements of any shape.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        shape: :class:`tuple`, optional
            Shape of the array to populate. By default,
            a shapeless array of one random element will be generated.
        src: sequence of :class:`int`, optional
            Players that will contribute to random array generation.  By default,
            all players contribute.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared array of random field elements.
        """

        if shape is None:
            shape=()

        if generator is None:
            generator = numpy.random.default_rng()

        return AdditiveArrayShare(self.field.uniform(size=shape, generator=generator))


    def floor(self, operand, *, encoding=None):
        """Privacy-preserving elementwise floor of encoded, secret-shared arrays.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared values to which floor should be applied.
        encoding: :class:`object`, optional
            Determines the number of fractional bits used for encoded values.
            The protocol's :attr:`encoding` is used by default if :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared floor of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        encoding = self._require_encoding(encoding)

        one = self.share(src=0, secret=self.field.full_like(operand.storage, 2**encoding.precision), shape=operand.storage.shape, encoding=Identity())
        shift_op = self.field.full_like(operand.storage, 2**encoding.precision)
        pl2 = self.field.full_like(operand.storage, self.field.order-1)

        abs_op = self.absolute(operand)
        frac_bits = encoding.precision
        field_bits = self.field.bits
        lsbs = self.bit_decompose(abs_op, bits=encoding.precision)
        lsbs_composed = self.bit_compose(lsbs)
        lsbs_inv = self.negative(lsbs_composed)
        two_lsbs = AdditiveArrayShare(self.field.multiply(lsbs_composed.storage, self.field.full_like(lsbs_composed.storage, 2)))
        ltz = self.less_zero(operand)
        ones2sub = AdditiveArrayShare(self.field.multiply(self.field_power(lsbs_composed, pl2).storage, shift_op))
        sel_2_lsbs = self.field_multiply(self.field_subtract(two_lsbs, ones2sub), ltz)
        return self.field_add(self.field_add(sel_2_lsbs, lsbs_inv), operand)


    def less(self, lhs, rhs):
        r"""Privacy-preserving elementwise less-than comparison.

        The result is the secret shared elementwise comparison :math:`lhs \lt rhs`.
        Note that the results will contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared values to be compared.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared values to be compared.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared comparison :math:`lhs \lt rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")

        one = self.field.full_like(lhs.storage, 1)
        two = self.field.full_like(lhs.storage, 2)
        twolhs = self.field_multiply(two, lhs)
        tworhs = self.field_multiply(two, rhs)
        diff = self.field_subtract(lhs, rhs)
        twodiff = self.field_multiply(two, diff)
        w = self.field_subtract(one, self._lsb(twolhs))
        x = self.field_subtract(one, self._lsb(tworhs))
        y = self.field_subtract(one, self._lsb(twodiff))
        wxorx = self.logical_xor(w,x)
        notwxorx = self.field_subtract(one, wxorx)
        xwxorx = self.field_multiply(x, wxorx)
        noty = self.field_subtract(one, y)
        notwxorxnoty = self.field_multiply(notwxorx, noty)
        return self.field_add(xwxorx, notwxorxnoty)


    def less_zero(self, operand):
        r"""Privacy-preserving elementwise less-than-zero comparison.

        The result is the secret shared elementwise comparison :math:`operand \lt 0`.
        Note that the results will contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared values to be compared.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared values to be compared.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared comparison :math:`operand \lt 0`.
        """
        self._assert_unary_compatible(operand, "operand")
        two = self.field.full_like(operand.storage, 2)
        result = self.field_multiply(two, operand)
        return self._lsb(result)


    def logical_and(self, lhs, rhs):
        """Privacy-preserving elementwise logical AND of secret shared arrays.

        The operands *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical AND of `lhs`
        and `rhs`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared array for logical AND.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared array for logical AND.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise logical AND of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return self.field_multiply(lhs, rhs)


    def logical_not(self, operand):
        """Privacy-preserving elementwise logical NOT.

        The operand *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical NOT of
        `operand`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared array for logical NOT.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise logical NOT of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return self.field_subtract(self.field.ones_like(operand.storage), operand)


    def logical_or(self, lhs, rhs):
        """Privacy-preserving elementwise logical OR of secret shared arrays.

        The operands *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical OR of `lhs`
        and `rhs`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared array for logical OR.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared array for logical OR.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise logical OR of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        total = self.field_add(lhs, rhs)
        product = self.field_multiply(lhs, rhs)
        return self.field_subtract(total, product)


    def logical_xor(self, lhs, rhs):
        """Privacy-preserving elementwise logical XOR of secret shared arrays.

        The operands *must* contain the field values :math:`0` or :math:`1`.
        The result will be the secret shared elementwise logical XOR of `lhs`
        and `rhs`, and will also contain the field values :math:`0` or
        :math:`1`, which do not require decoding if revealed.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared array for logical XOR.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared array for logical XOR.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise logical XOR of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        total = self.field_add(lhs, rhs)
        product = self.field_multiply(lhs, rhs)
        twice_product = self.field_multiply(self.field(2), product)
        return self.field_subtract(total, twice_product)


    def _lsb(self, operand):
        """Return the elementwise least significant bit of a secret shared array.

        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared array from which the least significant bits will be extracted

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Additive shared array containing the elementwise least significant
            bits of `operand`.
        """
        one = numpy.array(1, dtype=self.field.dtype)
        lop = AdditiveArrayShare(operand.storage.flatten())
        tmpBW, tmp = self.random_bitwise_secret(bits=self.field.bits, shape=lop.storage.shape)
        maskedlop = self.field_add(lop, tmp)
        c = self.reveal(maskedlop, encoding=Identity())
        comp_result = self._public_bitwise_less_than(lhspub=c, rhs=tmpBW)
        c = (c % 2)
        c0xr0 = numpy.empty(c.shape, dtype = self.field.dtype)
        for i, lc in enumerate(c):
            if lc:
                c0xr0[i] = self.field_subtract(lhs=one, rhs=AdditiveArrayShare(numpy.array(tmpBW.storage[i][-1], dtype=self.field.dtype))).storage
            else:
                c0xr0[i] = tmpBW.storage[i][-1]
        c0xr0 = AdditiveArrayShare(c0xr0)
        result = self.field_multiply(lhs=comp_result, rhs=c0xr0)
        result = AdditiveArrayShare(self.field.multiply(lhs=self.field.full_like(result.storage, 2), rhs=result.storage))
        result = self.field_subtract(lhs=c0xr0, rhs=result)
        result = self.field_add(lhs=comp_result, rhs=result)
        return AdditiveArrayShare(result.storage.reshape(operand.storage.shape))


    def maximum(self, lhs, rhs):
        """Privacy-preserving elementwise maximum of secret shared arrays.

        The result is the secret shared elementwise maximum of the operands.
        Note: the magnitude of the field elements should be less than one
        quarter of the field order for this method to be accurate in general.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared operand.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared operand.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise maximum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        max_share = self.field_add(self.field_add(lhs, rhs), self.absolute(self.field_subtract(lhs, rhs)))
        shift_right = self.field.full_like(lhs.storage, pow(2, self.field.order-2, self.field.order))
        max_share = self.field_multiply(max_share, shift_right)
        return max_share


    def minimum(self, lhs, rhs):
        """Privacy-preserving elementwise minimum of secret shared arrays.

        The result is the secret shared elementwise minimum of the operands.
        Note: the magnitude of the field elements should be less than one
        quarter of the field order for this method to be accurate in general.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared operand.
        rhs: :class:`AdditiveArrayShare`, required
            Secret shared operand.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise minimum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        diff = self.field_subtract(lhs, rhs)
        abs_diff = self.absolute(diff)
        min_share = self.field_subtract(self.field_add(lhs, rhs), abs_diff)
        shift_right = self.field.full_like(lhs.storage, pow(2, self.field.order-2, self.field.order))
        min_share = self.field_multiply(min_share, shift_right)
        return min_share


    def multiplicative_inverse(self, operand):
        """Privacy-preserving elementwise multiplicative inverse of a secret shared array.

        Returns the multiplicative inverse of a secret shared array
        in the context of the underlying finite field. Explicitly, this
        function returns a same shape array which, when multiplied
        elementwise with `operand`, will return a same shape array comprised
        entirely of ones, assuming `operand` is entirely non-trivial elements.

        This function does not take into account any field-external symantics.
        There is a potential for information leak here if `operand` contains any
        zero elements, that will be revealed. There is a small probability,
        2^-operand.storage.size, for this approach to fail by zero being randomly
        generated by the parties as the mask.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared operand to be multiplicatively inverted.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise multiplicative inverse of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")

        mask = self.field_uniform(shape=operand.storage.shape)
        masked_op = self.field_multiply(mask, operand)
        revealed_masked_op = self.reveal(masked_op, encoding=Identity())
        nppowmod = numpy.vectorize(lambda a, b, c: pow(int(a), int(b), int(c)), otypes=[self.field.dtype])
        inv = numpy.array(nppowmod(revealed_masked_op, self.field.order-2, self.field.order), dtype=self.field.dtype)
        op_inv_share = self.field.multiply(inv, mask.storage)
        return AdditiveArrayShare(op_inv_share)


    def multiply(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise multiplication of arrays.

        This method can be used to perform private-private, public-private, and
        private-public multiplication.  The result is the secret shared
        elementwise sum of the operands.  Note that public-public
        multiplication isn't allowed, as it isn't privacy-preserving!

        Unlike :meth:`field_multiply`, :meth:`multiply` is encoding aware:
        encoding is performed on its public inputs, and the results are
        shifted right to produce correct results when decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public value to be multiplied.
        encoding: :class:`object`, optional
            Encodes public operands and determines the number of bits to shift
            right the results.  The protocol's :attr:`encoding` is used by
            default if :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared product of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private multiplication.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            result = self.field_multiply(lhs, rhs)
            result = self.right_shift(result, bits=encoding.precision)
            return result

        # Private-public multiplication.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            result = self.field_multiply(lhs, encoding.encode(rhs, self.field))
            result = self.right_shift(result, bits=encoding.precision)
            return result

        # Public-private multiplication.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, AdditiveArrayShare):
            result = self.field_multiply(encoding.encode(lhs, self.field), rhs)
            result = self.right_shift(result, bits=encoding.precision)
            return result

        raise NotImplementedError(f"Privacy-preserving multiplication not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def negative(self, operand):
        """Privacy-preserving elementwise additive inverse of a secret shared array.

        Returns the additive inverse of a secret shared array
        in the context of the underlying finite field. Explicitly, this
        function returns a same shape array which, when added
        elementwise with `operand`, will return a same shape array comprised
        entirely of zeros.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared operand to be additively inverted.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise additive inverse of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return self.field_subtract(self.field.full_like(operand.storage, self.field.order), operand)


#    def pade_approx(self, func, operand, *, encoding=None, center=0, degree=12, scale=3):
#        """Return the pade approximation of `func` sampled with `operand`.
#
#        Note
#        ----
#        This is a collective operation that *must* be called
#        by all players that are members of :attr:`communicator`.
#
#        Parameters
#        ----------
#        func: callable object, required
#            The function to be approximated via the pade method.
#        operand: :class:`AdditiveArrayShare`, required
#            Secret-shared values where `func` should be evaluated.
#        center: :class:`float`, optional
#            The value at which the approximation should be centered. Sample
#            errors will be larger the further they are from this point.
#
#        Returns
#        -------
#        result: :class:`AdditiveArrayShare`
#            Secret shared result of evaluating the pade approximant of func(operand) with the given parameters.
#        """
#        from scipy.interpolate import approximate_taylor_polynomial, pade
#        num_deg = degree%2+degree//2
#        den_deg = degree//2
#
#        self._assert_unary_compatible(operand, "operand")
#        encoding = self._require_encoding(encoding)
#
#        func_taylor = approximate_taylor_polynomial(func, center, degree, scale)
#        func_pade_num, func_pade_den = pade([x for x in func_taylor][::-1], den_deg, n=num_deg)
#        enc_func_pade_num = encoding.encode(numpy.array([x for x in func_pade_num]), self.field)
#        enc_func_pade_den = encoding.encode(numpy.array([x for x in func_pade_den]), self.field)
#
#        result_list=[]
#        for op in operand.storage:
#            single_op_share = AdditiveArrayShare(numpy.array(op, dtype=object))
#            op_pows_num_list = [self.share(src=1, secret=numpy.array(1), shape=())]
#            for i in range(num_deg):
#                op_pows_num_list.append(self.multiply(single_op_share, op_pows_num_list[-1]))
#            if degree%2:
#                op_pows_den_list=[thing for thing in op_pows_num_list[:-1]]
#            else:
#                op_pows_den_list=[thing for thing in op_pows_num_list]
#            op_pows_num = AdditiveArrayShare(numpy.array([x.storage for x in op_pows_num_list]))
#            op_pows_den = AdditiveArrayShare(numpy.array([x.storage for x in op_pows_den_list]))
#
#            result_num_prod = self.field_multiply(op_pows_num, enc_func_pade_num)
#            result_num = self.right_shift(self.sum(result_num_prod), bits=encoding.precision)
#
#            result_den_prod = self.field_multiply(op_pows_den, enc_func_pade_den)
#            result_den = self.right_shift(self.sum(result_den_prod), bits=encoding.precision)
#            result_list.append(self.divide(result_num, result_den))
#        return AdditiveArrayShare(numpy.array([s.storage for s in result_list], dtype=object))


    def power(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise exponentiation.

        Raises secret shared array values to public integer values.  Unlike
        :meth:`field_power`, :meth:`power` operates on encoded values, shifting
        the results right to ensure correct decoded results.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare`, required
            Secret shared values which iwll be raised to a power.
        rhs: :class:`int` or integer :class:`numpy.ndarray`, required
            Public integer power(s) to which each element in `lhs` will be raised.
        encoding: :class:`object`, optional
            Determines the number of bits to shift right the results.  The
            protocol's :attr:`encoding` is used by default if :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared result of raising `lhs` to the power(s) in `rhs`.
        """
        encoding = self._require_encoding(encoding)

        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, (numpy.ndarray, int)):
            if isinstance(rhs, int):
                rhs = self.field.full_like(lhs.storage, rhs)

            ans=[]
            for lhse, rhse in numpy.nditer([lhs.storage, rhs], flags=(["refs_ok"])):
                rhsbits = [int(x) for x in bin(rhse)[2:]][::-1]
                tmp = AdditiveArrayShare(lhse)
                it_ans = self.share(src = 0, secret=self.field.full_like(tmp.storage, encoding.encode(numpy.array(1), self.field)),shape=tmp.storage.shape, encoding=Identity())
                limit = len(rhsbits)-1
                for i, bit in enumerate(rhsbits):
                    if bit:
                        it_ans = self.field_multiply(it_ans, tmp)
                        it_ans = self.right_shift(it_ans, bits=encoding.precision)
                    if i < limit:
                        tmp = self.field_multiply(tmp,tmp)
                        tmp = self.right_shift(tmp, bits=encoding.precision)
                ans.append(it_ans)
            return AdditiveArrayShare(numpy.array([x.storage for x in ans], dtype=self.field.dtype).reshape(lhs.storage.shape))

        raise NotImplementedError(f"Privacy-preserving exponentiation not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def _public_bitwise_less_than(self, *, lhspub, rhs):
        """Comparison Operator

        Parameters
        ----------
        lhs: :class:`ndarray`, required
            a publically known numpy array of integers and one of the two objects to be compared
        rhs: :class:`AdditiveArrayShare`, required
            bit decomposed shared secret(s) and the other of the two objects to be compared
            the bitwidth of each value in rhs (deduced from its shape) is taken to be the
            bitwidth of interest for the comparison if the values in lhspub require more bits
            for their representation, the computation will be incorrect

        note: this method is private as it does not consider the semantic mapping of meaning
        onto the field. The practical result of this is that every negative value will register as
        greater than every positive value due to the encoding.


        Returns
        -------
        an additive shared array containing the element wise result of the comparison: result[i] = 1 if lhspub[i] < rhs[i] and 0 otherwise
        """
        if lhspub.shape != rhs.storage.shape[:-1]:
            raise ValueError('rhs is not of the expected shape - it should be the same as lhs except the last dimension') # pragma: no cover
        bitwidth = rhs.storage.shape[-1]
        lhsbits = []
        for val in lhspub:
            tmplist = [int(x) for x in bin(val)[2:]]
            if len(tmplist) < bitwidth:
                tmplist = [0 for x in range(bitwidth-len(tmplist))] + tmplist
            lhsbits.append(tmplist)
        lhsbits = numpy.array(lhsbits, dtype=self.field.dtype)
        assert(lhsbits.shape == rhs.storage.shape)
        one = numpy.array(1, dtype=self.field.dtype)
        flatlhsbits = lhsbits.reshape((-1, lhsbits.shape[-1]))
        flatrhsbits = rhs.storage.reshape((-1, rhs.storage.shape[-1]))
        results=[]
        for j in range(len(flatlhsbits)):
            xord = []
            preord = []
            msbdiff=[]
            rhs_bit_at_msb_diff = []
            for i in range(bitwidth):
                rhsbit=AdditiveArrayShare(storage=numpy.array(flatrhsbits[j,i], dtype=self.field.dtype))
                if flatlhsbits[j][i] == 1:
                    xord.append(self.field_subtract(lhs=one, rhs=rhsbit))
                else:
                    xord.append(rhsbit)
            preord = [xord[0]]
            for i in range(1,bitwidth):
                preord.append(self.logical_or(lhs=preord[i-1], rhs=xord[i]))
            msbdiff = [preord[0]]
            for i in range(1,bitwidth):
                msbdiff.append(self.field_subtract(lhs=preord[i], rhs=preord[i-1]))
            for i in range(bitwidth):
                rhsbit=AdditiveArrayShare(storage=numpy.array(flatrhsbits[j,i], dtype=self.field.dtype))
                rhs_bit_at_msb_diff.append(self.field_multiply(rhsbit, msbdiff[i]))
            result = rhs_bit_at_msb_diff[0]
            for i in range(1,bitwidth):
                result = self.field_add(lhs=result, rhs=rhs_bit_at_msb_diff[i])
            results.append(result)
        return AdditiveArrayShare(numpy.array([x.storage for x in results], dtype=self.field.dtype).reshape(rhs.storage.shape[:-1]))


    def random_bitwise_secret(self, *, bits, shape=None, src=None, generator=None):
        """Return secret values created by combining randomly generated bits.

        This method returns two outputs - a secret shared array of randomly
        generated bits, and a secret shared array of values created by
        combining the bits in big-endian order.  It is secure against
        non-colluding semi-honest adversaries.  A subset of players (by
        default: all) generate and secret share vectors of pseudo-random bits
        which are then XOR-ed together elementwise.  Communication and
        computation costs increase with the number of bits and the number of
        players, while security increases with the number of players.

        The bit array will have one more dimension than the secret array,
        containing the bits in big-endian order.

        .. warning::

             If you supply your own generators, be careful to ensure that each
             has a unique seed value to preserve privacy (for example: a
             constant plus the player's rank).  If players receive generators
             with identical seed values, even numbers of players will produce
             all zero bits.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`, even
        if they aren't participating in the random bit generation.

        Parameters
        ----------
        bits: :class:`int`, required
            Number of bits to generate.
        shape: sequence of :class:`int`, optional
            Shape of the output `secrets` array.  The output `bits` array will have this
            shape, plus one extra dimension for the bits.
        src: sequence of :class:`int`, optional
            Players that will contribute to random bit generation.  By default,
            all players contribute.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        bits: :class:`AdditiveArrayShare`
            Secret shared array of randomly-generated bits, with shape
            :math:`shape \\times bits`.
        secrets: :class:`AdditiveArrayShare`
            Secret shared array of values created by combining the generated
            bits in big-endian order, with shape `shape`.
        """
        bits = int(bits)
        shape_was_none = False
        if bits < 1:
            raise ValueError(f"bits must be a positive integer, got {bits} instead.") # pragma: no cover

        if src is None:
            src = self.communicator.ranks
        if not src:
            raise ValueError(f"src must include at least one player, got {src} instead.") # pragma: no cover

        if generator is None:
            generator = numpy.random.default_rng()
        if shape is None:
            shape = ()
            shape_was_none = True
        bit_res = []
        share_res = []
        numzeros = numpy.zeros(shape)
        for loopop in numzeros.flat:
            # Each participating player generates a vector of random bits.
            if self.communicator.rank in src:
                local_bits = generator.choice(2, size=bits).astype(self.field.dtype)
            else:
                local_bits = None

            # Each participating player secret shares their bit vectors.
            player_bit_shares = []
            for rank in src:
                player_bit_shares.append(self.share(src=rank, secret=local_bits, shape=(bits,), encoding=Identity()))

            # Generate the final bit vector by xor-ing everything together elementwise.
            bit_share = player_bit_shares[0]
            for player_bit_share in player_bit_shares[1:]:
                bit_share = self.logical_xor(bit_share, player_bit_share)

            # Shift and combine the resulting bits in big-endian order to produce a random value.
            shift = numpy.power(2, numpy.arange(bits, dtype=self.field.dtype)[::-1])
            shifted = self.field.multiply(shift, bit_share.storage)
            secret_share = AdditiveArrayShare(numpy.array(numpy.sum(shifted), dtype=self.field.dtype))
            bit_res.append(bit_share)
            share_res.append(secret_share)
        if shape_was_none:
            bit_share = AdditiveArrayShare(numpy.array([x.storage for x in bit_res], dtype=self.field.dtype).reshape(bits))
            secret_share = AdditiveArrayShare(numpy.array([x.storage for x in share_res], dtype=self.field.dtype).reshape(shape))#, order="C"))
        else:
            bit_share = AdditiveArrayShare(numpy.array([x.storage for x in bit_res], dtype=self.field.dtype).reshape(shape+(bits,)))#, order="C"))
            secret_share = AdditiveArrayShare(numpy.array([x.storage for x in share_res], dtype=self.field.dtype).reshape(shape))#, order="C"))

        return bit_share, secret_share


    def relu(self, operand):
        """Privacy-preserving elementwise ReLU of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared operand to which the ReLU function will be applied.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise ReLU of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        ltz = self.less_zero(operand)
        nltz = self.logical_not(ltz)
        nltz_parts = self.field_multiply(nltz, operand)
        return nltz_parts


    def reshare(self, operand):
        """Privacy-preserving re-randomization of a secret shared array.

        This method returns a new set of secret shares that contain different
        random values than the input but represent the same secret value.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared operand which should be re-randomized.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared, re-randomized version of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        recshares = []
        for i in range(self.communicator.world_size):
            recshares.append(self.share(src=i, secret=operand.storage, shape=operand.storage.shape, encoding=Identity()))
        acc = numpy.zeros(operand.storage.shape, dtype=self.field.dtype)
        for s in recshares:
            acc += s.storage
        acc %= self.field.order
        return AdditiveArrayShare(acc)


    def reveal(self, share, *, dst=None, encoding=None):
        """Reveals a secret shared value to a subset of players.

        Note
        ----
        This is a collective operation that *must* be called by all players
        that are members of :attr:`communicator`, whether they are receiving
        the revealed secret or not.

        Parameters
        ----------
        share: :class:`AdditiveArrayShare`, required
            The local share of the secret to be revealed.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.
        encoding: :class:`object`, optional
            Encoding used to extract the revealed secret from field values. The
            protocol's default encoding will be used if `None`.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            The revealed secret, if this player is a member of `dst`, or :any:`None`.
        """
        if not isinstance(share, AdditiveArrayShare):
            raise ValueError("share must be an instance of AdditiveArrayShare.") # pragma: no cover

        # Identify who will be receiving shares.
        if dst is None:
            dst = self.communicator.ranks

        encoding = self._require_encoding(encoding)

        # Send data to the other players.
        secret = None
        for recipient in dst:
            received_shares = self.communicator.gather(value=share.storage, dst=recipient)

            # If we're a recipient, recover the secret.
            if self.communicator.rank == recipient:
                secret = received_shares[0].copy()
                for received_share in received_shares[1:]:
                    self.field.inplace_add(secret, received_share)

        return encoding.decode(secret, self.field)


    def right_shift(self, operand, *, bits, src=None, generator=None, trunc_mask=None, rem_mask=None):
        """Privacy-preserving elementwise right-shift of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Also, this approach may probabilstically introduce a small error in the least significant bits 
        of the result. The is that we generate masks for the two sections of the secret shared value 
        of interest, one to be used in clearing out the low order bits, the other for masking the 
        higher order bits that will remain. There is a chance that adding the masks in will generate 
        a carry into the region that will remain after the shift. Therefore there is a chance that the 
        effects of that carry will remain in the least significant bits of the final result, which, 
        with the default parameters and encoding, means there could be an error in the final result on 
        the order of 2^-16, and with decreasing probability 2^-15, or 2^-14 depending on how far the 
        carry propagates.


        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret-shared values to be shifted right.
        bits: :class:`int`, optional
            Number of bits to shift.
        src: sequence of :class:`int`, optional
            Players who will participate in generating random bits as part of
            the shift process.  More players increases security but
            decreases performance.  Defaults to all players.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for generating random bits.  By
            default, `numpy.random.default_rng()` will be used.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared result of shifting `operand` to the right by `bits` bits.
        """
        if not isinstance(operand, AdditiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of AdditiveArrayShare, got {type(operand)} instead.") # pragma: no cover

        shift_left = self.field.full_like(operand.storage, 2**bits)
        # Multiplicative inverse of shift_left.
        shift_right = self.field.full_like(operand.storage, pow(2**bits, self.field.order-2, self.field.order))

        if trunc_mask:
            truncation_mask = trunc_mask
        else:
            # Generate random bits that will mask the region to be truncated.
            _, truncation_mask = self.random_bitwise_secret(bits=bits, src=src, generator=generator, shape=operand.storage.shape)
        if rem_mask:
            remaining_mask = rem_mask
        else:
            # Generate random bits that will mask everything outside the region to be truncated.
            _, remaining_mask = self.random_bitwise_secret(bits=self.field.bits-bits, src=src, generator=generator, shape=operand.storage.shape)
        remaining_mask.storage = self.field.multiply(remaining_mask.storage, shift_left)

        # Combine the two masks.
        mask = self.field_add(remaining_mask, truncation_mask)

        # Mask the array element.
        masked_element = self.field_add(mask, operand)

        # Reveal the element to all players (because it's masked, no player learns the underlying secret).
        masked_element = self.reveal(masked_element, encoding=Identity())

        # Retain just the bits within the region to be truncated, which need to be removed.
        masked_truncation_bits = numpy.array(masked_element % shift_left, dtype=self.field.dtype)

        # Remove the mask, leaving just the bits to be removed from the
        # truncation region.  Because the result of the subtraction is
        # secret shared, the secret still isn't revealed.
        truncation_bits = self.field_subtract(masked_truncation_bits, truncation_mask)

        # Remove the bits in the truncation region from the element.  The result can be safely truncated.
        result = self.field_subtract(operand, truncation_bits)

        # Truncate the element by shifting right to get rid of the (now cleared) bits in the truncation region.
        result = self.field.multiply(result.storage, shift_right)

        return AdditiveArrayShare(result)


    def share(self, *, src, secret, shape, encoding=None):
        """Convert an array of scalars to an additive secret share.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        src: :class:`int`, required
            The player providing the private array to be secret shared.
        secret: :class:`numpy.ndarray` or :any:`None`, required
            The secret array to be shared.  This value is ignored for all
            players except `src`.
        shape: :class:`tuple`, required
            The shape of the secret.  Note that all players must specify the
            same shape.
        encoding: :class:`object`, optional
            Encoding used to convert `secret` into field values. The protocol's
            default encoding will be used if `None`.

        Returns
        -------
        share: :class:`AdditiveArrayShare`
            The local share of the secret shared array.
        """
        if not isinstance(shape, tuple):
            shape = (shape,)

        if self.communicator.rank == src:
            if not isinstance(secret, numpy.ndarray):
                raise ValueError("secret must be an instance of numpy.ndarray.") # pragma: no cover
            if secret.shape != shape:
                raise ValueError(f"Expected secret.shape {shape}, got {secret.shape} instead.") # pragma: no cover

        encoding = self._require_encoding(encoding)

        # Generate a pseudorandom zero-sharing.
        przs = self._przs(shape=shape)

        # Add the secret to the PRZS
        if self.communicator.rank == src:
            secret = encoding.encode(secret, self.field)
            self.field.inplace_add(przs, secret)
        # Package the result.
        return AdditiveArrayShare(przs)


    def subtract(self, lhs, rhs, *, encoding=None):
        """Privacy-preserving elementwise difference of arrays.

        This method can be used to perform private-private, public-private, and
        private-public addition.  The result is the secret shared elementwise
        difference of the operands.  Note that public-public subtraction isn't
        allowed, as it isn't privacy-preserving!

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be subtracted.
        rhs: :class:`AdditiveArrayShare` or :class:`numpy.ndarray`, required
            Secret shared or public values to be subtracted.
        encoding: :class:`object`, optional
            Encodes public operands.  The protocol's :attr:`encoding`
            is used by default if :any:`None`.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared difference of `lhs` and `rhs`.
        """
        encoding = self._require_encoding(encoding)

        # Private-private subtraction.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, AdditiveArrayShare):
            return self.field_subtract(lhs, rhs)

        # Private-public subtraction.
        if isinstance(lhs, AdditiveArrayShare) and isinstance(rhs, numpy.ndarray):
            return self.field_subtract(lhs, encoding.encode(rhs, self.field))

        # Public-private subtraction.
        if isinstance(lhs, numpy.ndarray) and isinstance(rhs, AdditiveArrayShare):
            return self.field_subtract(encoding.encode(lhs, self.field), rhs)

        raise NotImplementedError(f"Privacy-preserving subtraction not implemented for the given types: {type(lhs)} and {type(rhs)}.") # pragma: no cover


    def sum(self, operand):
        """Privacy-preserving sum of a secret shared array's elements.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared array containing elements to be summed.

        Returns
        -------
        value: :class:`AdditiveArrayShare`
            Secret-shared sum of `operand`'s elements.
        """
        self._assert_unary_compatible(operand, "operand")
        return AdditiveArrayShare(self.field.sum(operand.storage))


#    def taylor_approx(self, func, operand, *, encoding=None, center=0, degree=7, scale=3):
#        """Return the taylor approximation of `func` sampled with `operand`.
#
#        Note
#        ----
#        This is a collective operation that *must* be called
#        by all players that are members of :attr:`communicator`.
#
#        Parameters
#        ----------
#        func: callable object, required
#            The function to be approximated via the taylor method
#        operand: :class:`AdditiveArrayShare`, required
#            Secret-shared values where `func` should be evaluated.
#        center: :class:`float`, optional
#            The value at which the approximation should be centered. Sample
#            errors will be larger the further they are from this point.
#
#        Returns
#        -------
#        result: :class:`AdditiveArrayShare`
#            Secret shared result of evaluating the taylor approximant of func(operand) with the given parameters
#        """
#        from scipy.interpolate import approximate_taylor_polynomial
#
#        self._assert_unary_compatible(operand, "operand")
#        encoding = self._require_encoding(encoding)
#
#        taylor_poly = approximate_taylor_polynomial(func, center, degree, scale)
#
#        enc_taylor_coef = encoding.encode(numpy.array([x for x in taylor_poly]), self.field)
#        result_list=[]
#        for op in operand.storage:
#            single_op_share = AdditiveArrayShare(numpy.array(op, dtype=object))
#            op_pow_list = [self.share(src=1, secret=numpy.array(1), shape=())]
#            for i in range(degree):
#                op_pow_list.append(self.multiply(single_op_share, op_pow_list[-1]))
#
#            op_pow_shares = AdditiveArrayShare(numpy.array([x.storage for x in op_pow_list]))
#
#            result = self.field_multiply(op_pow_shares, enc_taylor_coef)
#            result = self.sum(result)
#            result = self.right_shift(result, bits=encoding.precision)
#            result_list.append(result)    
#        return AdditiveArrayShare(numpy.array([s.storage for s in result_list], dtype=object))


    def zigmoid(self, operand, *, encoding=None):
        r"""Privacy-preserving elementwise zigmoid of a secret shared array.

        Zigmoid is a piecewise approximation to the popular sigmoid activation
        function that is more efficient to compute in an MPC context:

        .. math::

            zigmoid(x) = \left\{
            \begin{array}\\
                0 & if\ x<-0.5 \\
                x+0.5 & if\ -0.5\leq x \leq 0.5 \\
                1 & if x>0.5
            \end{array}
            \right.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`AdditiveArrayShare`, required
            Secret shared operand to which the zigmoid function will be applied.

        Returns
        -------
        result: :class:`AdditiveArrayShare`
            Secret-shared elementwise zigmoid of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        encoding = self._require_encoding(encoding)

        ones = encoding.encode(numpy.full_like(operand.storage, 1.0), self.field)
        half = encoding.encode(numpy.full_like(operand.storage, 0.5), self.field)

        secret_plushalf = self.field_add(half, operand)
        secret_minushalf = self.field_subtract(operand, half)
        ltzsmh = self.less_zero(secret_minushalf)
        nltzsmh = self.logical_not(ltzsmh)
        ltzsph = self.less_zero(secret_plushalf)
        middlins = self.field_subtract(ltzsmh, ltzsph)
        extracted_middlins = self.field_multiply(middlins, operand)
        extracted_halfs = self.field_multiply(middlins, half)
        extracted_middlins = self.field_add(extracted_middlins, extracted_halfs)
        ones_part = self.field_multiply(nltzsmh, ones)
        return self.field_add(ones_part, extracted_middlins)




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for user interaction."""

import numpy

from cicada.communicator.interface import Communicator

def secret_input(*, communicator, src, prompt=None, dtype=float, timeout=300):
    """Prompt one user for a secret.

    Note
    ----
    Although this function only prompts one player for input, it is a
    collective operation that *must* be called by all players that are members
    of `communicator`.

    Parameters
    ----------
    communicator: :class:`~cicada.communicator.interface.Communicator`, required
        Used to coordinate among players.
    src: :class:`int`, required
        Rank of the player who will be prompted for a secret.
    prompt: :class:`str`, optional
        Override the default interactive prompt.  See :func:`input`
        for usage.
    dtype: :func:`callable`, optional
        Function for parsing user input into a final value.  The function must
        take one :class:`str` argument and return a result.  The builtin
        functions :class:`int`, :class:`float`, and :class:`str` are useful
        examples.
    timeout: :class:`numbers.Number`, optional
        Maximum time to wait for user input, in seconds. Defaults to 300 seconds.

    Returns
    -------
    value: :class:`object` or :any:`None`
        For player `src`: the secret input.  For all other players: :any:`None`.
    """
    if communicator.rank == src:
        if prompt is None:
            prompt = f"Enter a secret: "
        secret = numpy.array(dtype(input(prompt)))
    else:
        secret = None

    # Wait for user input.
    with communicator.override(timeout=timeout):
        communicator.barrier()

    return secret



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for encoding domain values using integer fields."""

import numbers

import numpy


class Bits(object):
    """Converts arrays of bit values to and from field values.
    """

    def __eq__(self, other):
        return isinstance(other, Bits)


    def __repr__(self):
        return f"cicada.encoding.Bits()" # pragma: no cover


    def decode(self, array, field):
        """Convert an array of field values to bit values.

        Parameters
        ----------
        array: :class:`numpy.ndarray`, or :any:`None`, required
            Array of field values containing only :math:`0` or :math:`1`.
        field: :class:`cicada.arithmetic.Field`, required
            Field over which the values in `array` are defined.

        Returns
        -------
        decoded: :class:`numpy.ndarray` or :any:`None`
            Array of integers containing the values :math:`0` and :math:`1`, or
            :any:`None` if the input was :any:`None`.

        Raises
        ------
        ValueError
            If `array` contains anything but :math:`0` or :math:`1`.
        """
        if array is None:
            return array
        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: no cover
        if not array.dtype == field.dtype:
            raise ValueError(f"Expected array dtype to be {field.dtype}, got {array.dtype} instead.") # pragma: no cover

        # Strict enforcement - input must only contain zeros and ones.
        result = array.astype(bool)
        if not numpy.array_equal(array, result):
            raise ValueError(f"Expected array to contain only zeros and ones, got {array} instead.") # pragma: no cover

        return array.astype(numpy.uint8)


    def encode(self, array, field):
        """Convert an array of bit values to field values.

        Parameters
        ----------
        array: :class:`numpy.ndarray` or :any:`None`, required
            Array to convert containing only :math:`0` or :math:`1`.
        field: :class:`cicada.arithmetic.Field`, required
            Field over which the returned values are defined.

        Returns
        -------
        encoded: :class:`numpy.ndarray` or :any:`None`
            Encoded array with the same shape as the input, containing the
            `field` values :math:`0` and :math:`1`, or :any:`None` if the input
            was :any:`None`.

        Raises
        ------
        ValueError
            If `array` contains anything but :math:`0` or :math:`1`.
        """
        if array is None:
            return array

        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: no cover

        # Strict enforcement - input must only contain zeros and ones.
        result = array.astype(bool)
        if not numpy.array_equal(array, result):
            raise ValueError(f"Expected array to contain only zeros and ones, got {array} instead.") # pragma: no cover

        # Convert to the field.
        return field(result)


class Boolean(object):
    """Converts arrays of boolean values to and from field values.
    """

    def __eq__(self, other):
        return isinstance(other, Boolean)


    def __repr__(self):
        return f"cicada.encoding.Boolean()" # pragma: no cover


    def decode(self, array, field):
        """Convert an array of field values to boolean values.

        Parameters
        ----------
        array: :class:`numpy.ndarray`, or :any:`None`, required
            Array of field values containing only :math:`0` or :math:`1`.
        field: :class:`cicada.arithmetic.Field`, required
            Field over which the values in `array` are defined.

        Returns
        -------
        decoded: :class:`numpy.ndarray` or :any:`None`
            Array of boolean values :math:`True` and :math:`False`, or
            :any:`None` if the input was :any:`None`.
        """
        if array is None:
            return array
        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: no cover
        if not array.dtype == field.dtype:
            raise ValueError(f"Expected array dtype to be {field.dtype}, got {array.dtype} instead.") # pragma: nocover

        return array.astype(bool)


    def encode(self, array, field):
        """Convert an array of boolean values to field values.

        Parameters
        ----------
        array: :class:`numpy.ndarray` or :any:`None`, required
            Array to convert.  Nonzero values are considered :math:`True`, zero values are considered :math:`False`.
        field: :class:`cicada.arithmetic.Field`, required
            Field over which the returned values are defined.

        Returns
        -------
        encoded: :class:`numpy.ndarray` or :any:`None`
            Encoded array with the same shape as the input, containing the
            `field` values :math:`0` and :math:`1`, or :any:`None` if the input
            was :any:`None`.
        """
        if array is None:
            return array

        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: nocover

        # Permissive coercion of truthy values.
        result = array.astype(bool)

        # Convert to the field.
        return field(result)


class FixedPoint(object):
    """Encodes real values in a field using a fixed-point representation.

    Encoded values are :class:`numpy.ndarray` instances containing Python
    integers, with `precision` bits reserved for encoding fractional digits.
    Decoded values will be :class:`numpy.ndarray` instances containining 64-bit
    floating point values.

    Parameters
    ----------
    precision: :class:`int`, optional
        The number of bits reserved to store fractions in encoded values.  Defaults
        to 16.
    """
    def __init__(self, precision=16):
        if not isinstance(precision, numbers.Integral):
            raise ValueError(f"Expected integer precision, got {type(precision)} instead.") # pragma: no cover
        if precision < 0:
            raise ValueError(f"Expected non-negative precision, got {precision} instead.") # pragma: no cover

        self._precision = precision
        self._scale = int(2**self._precision)


    def __eq__(self, other):
        return isinstance(other, FixedPoint) and self._precision == other._precision


    def __repr__(self):
        return f"cicada.encoding.FixedPoint(precision={self._precision})" # pragma: no cover


    def decode(self, array, field):
        """Convert an array of field values to an array of real values.

        Parameters
        ----------
        array: :class:`numpy.ndarray`, or :any:`None`, required
            Array of field values created with :meth:`encode`.
        field: :class:`cicada.arithmetic.Field`, required
            Field used to create `array`.

        Returns
        -------
        decoded: :class:`numpy.ndarray`
            A floating point array with the same shape as the input, containing
            the decoded representation of `array`, or :any:`None` if the input
            was :any:`None`.
        """
        if array is None:
            return array

        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: nocover
        if not array.dtype == field.dtype:
            raise ValueError(f"Expected array dtype to be {field.dtype}, got {array.dtype} instead.") # pragma: nocover

        order = field.order
        posbound = order // 2

        # Set aside storage for the result (ensures that we return an array and not a scalar).
        output = numpy.empty_like(array, dtype=numpy.float64)
        # Convert from the field to a plain array of integers.
        result = numpy.copy(array, subok=False)
        # Switch from twos-complement notation to negative values.
        result = numpy.where(result > posbound, -(order - result), result)
        # Shift values back to the right and convert to reals.
        return numpy.divide(result.astype(numpy.float64), self._scale, out=output)


    def encode(self, array, field):
        """Convert array of real values to an array of field values using a fixed point integer representation.

        Parameters
        ----------
        array: :class:`numpy.ndarray` or :any:`None`, required
            The array to convert.

        field: :class:`cicada.arithmetic.Field`, required
            The returned array elements will be members of this field.

        Returns
        -------
        encoded: :class:`numpy.ndarray` or :any:`None`
            Encoded array with the same shape as the input, containing the
            fixed precision integer representation of `array`, or :any:`None`
            if the input was :any:`None`.
        """
        if array is None:
            return array

        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: nocover

        order = field.order
        posbound = order // 2

        # Ensure we have an array, but only copy data if necessary.
        result = numpy.asarray(array, dtype=numpy.float64)
        # Shift array values left.  Don't do this inline!
        result = result * self._scale
        # Test to be sure our values are in-range for the field.
        if numpy.any(numpy.abs(result) >= posbound):
            raise ValueError("Values to be encoded are too large for representation in the field.") # pragma: no cover
        # Convert to integers, using the Python modulo operator to handle negative values.
        result = numpy.array([int(x) % order for x in numpy.nditer(result)], dtype=field.dtype).reshape(result.shape)
        # Convert to a field.
        return field(result)


    @property
    def precision(self):
        return self._precision


class Identity(object):
    """Encodes and decodes field values without modification.

    Encoded values are :class:`numpy.ndarray` instances containing Python
    integers.  Decoded values will be the same.

    """

    def __eq__(self, other):
        return isinstance(other, Identity)


    def __repr__(self):
        return f"cicada.encoding.Identity()" # pragma: no cover


    def decode(self, array, field):
        """Return an array of field values without modification.

        Parameters
        ----------
        array: :class:`numpy.ndarray`, or :any:`None`, required
            Array of field values created with :meth:`encode`.

        Returns
        -------
        decoded: :class:`numpy.ndarray` or :any:`None`
            An array of Python integers, or :any:`None` if the input was
            :any:`None`.
        """
        if array is None:
            return array
        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: nocover
        if not array.dtype == field.dtype:
            raise ValueError(f"Expected array dtype to be {field.dtype}, got {array.dtype} instead.") # pragma: nocover
        return array


    def encode(self, array, field):
        """Convert array of integer values to an array of field values without modification.

        Parameters
        ----------
        array: :class:`numpy.ndarray` or :any:`None`, required
            The array to convert.

        field: :class:`cicada.arithmetic.Field`, required
            The returned array elements will be members of this field.

        Returns
        -------
        encoded: :class:`numpy.ndarray` or :any:`None`
            Encoded array with the same shape as the input, containing the
            fixed precision integer representation of `array`, or :any:`None`
            if the input was :any:`None`.
        """
        if array is None:
            return array

        if not isinstance(array, numpy.ndarray):
            raise ValueError(f"Expected array to be an instance of numpy.ndarray, got {type(array)} instead.") # pragma: nocover

        # Convert to a field.
        return field(array)



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

# Repackaging existing tests for PEP-8.

def assert_almost_equal(first, second, places=7, delta=None, msg=None):
    return unittest.TestCase().assertAlmostEqual(first, second, places=places, msg=msg, delta=delta)

def assert_dict_equal(first, second, msg=None):
    return unittest.TestCase().assertDictEqual(first, second, msg)

def assert_equal(first, second, msg=None):
    return unittest.TestCase().assertEqual(first, second, msg)

def assert_false(expr, msg=None):
    return unittest.TestCase().assertFalse(expr, msg)

def assert_is(first, second, msg=None):
    return unittest.TestCase().assertIs(first, second, msg)

def assert_is_instance(obj, cls, msg=None):
    return unittest.TestCase().assertIsInstance(obj, cls, msg)

def assert_is_none(expr, msg=None):
    return unittest.TestCase().assertIsNone(expr, msg)

def assert_logs(logger=None, level=None):
    return unittest.TestCase().assertLogs(logger, level)

def assert_no_logs(logger=None, level=None):
    return unittest.TestCase().assertNoLogs(logger, level)

def assert_sequence_equal(first, second, msg=None, seq_type=None):
    return unittest.TestCase().assertSequenceEqual(first, second, msg, seq_type)

def assert_true(expr, msg=None):
    return unittest.TestCase().assertTrue(expr, msg)



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pseudorandom Zero-Sharing functionality."""

import inspect

import numpy

from cicada.arithmetic import Field
from cicada.communicator.interface import Communicator, Tag


class PRZSProtocol(object):
    """Implements the Pseudorandom Zero-Sharing protocol of Cramer, Damgard, and Ishai.

    Note
    ----
    Creating the protocol is a collective operation that *must*
    be implemented by all players that are members of `communicator`.

    Parameters
    ----------
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that this protocol will use for communication.
    field: :class:`cicada.arithmetic.Field`, required
        The field from which generated values will be drawn.
    seed: :class:`int`, optional
        Seed used to initialize random number generators.  For privacy, this
        value should be different for each player.
    """
    def __init__(self, *, communicator, field, seed):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover
        self._communicator = communicator

        if not isinstance(field, Field):
            raise ValueError("A Cicada field is required.") # pragma: no cover
        self._field = field

        # Send random seed to next party, receive random seed from prev party
        if communicator.world_size >= 2:  # Otherwise sending seeds will segfault.
            next_rank = (communicator.rank + 1) % communicator.world_size
            prev_rank = (communicator.rank - 1) % communicator.world_size

            request = communicator.isend(value=seed, dst=next_rank, tag=Tag.PRZS)
            result = communicator.irecv(src=prev_rank, tag=Tag.PRZS)

            request.wait()
            result.wait()

            prev_seed = result.value
        else:
            prev_seed = seed

        # Setup random number generators
        self._g0 = numpy.random.default_rng(seed=seed)
        self._g1 = numpy.random.default_rng(seed=prev_seed)


    def __call__(self, *, shape):
        """Generate an array of field values that sum to zero, using pseudorandom zero-sharing.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        shape: :class:`tuple`, required
            The shape of the result.  Note that all players must specify the
            same shape.

        Returns
        -------
        przs: :class:`numpy.ndarray`
            The local share of zeros.
        """
        if not isinstance(shape, tuple):
            shape = (shape,)

        przs = self.field.uniform(size=shape, generator=self._g0)
        self.field.inplace_subtract(przs, self.field.uniform(size=shape, generator=self._g1))

        return przs


    @property
    def communicator(self):
        """The :class:`~cicada.communicator.interface.Communicator` used by this protocol."""
        return self._communicator # pragma: no cover


    @property
    def field(self):
        """The :class:`~cicada.arithmetic.Field` used by this protocol."""
        return self._field





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pprint

import numpy

import cicada.communicator
import cicada.encoder
import cicada.additive
import hmac
import hashlib
from time import time

logging.basicConfig(level=logging.INFO)

def main(communicator):
    B = 2000 #bloom filter size
    k = 14 # number of hash functions
    n = 100 # of elements in the filter
    log = cicada.Logger(logging.getLogger(), communicator)
    protocol = cicada.additive.AdditiveProtocolSuite(communicator)

    keys = [x for x in range(k)]
    hashes = []
    bf = [0 for x in range(B)]
    items = [i for i in range(n)]

    #build bloom filter
    t0=time()
    for i in items:
        for k in keys:
            bf[hash((k,i)) %B]=1

    log.info(f'time to construct in the clear: {time()-t0}s', src=0)
    #log.info(bf, src=0)

    #confirm bloom filter correctness
    t0=time()
    acc_list = [0 for x in range(n)]
    for i,e in enumerate(items):
        for k in keys:
            if bf[hash((k,e))%B]==1:
                acc_list[i] += 1
    log.info(f'time to make 100 centralized/clear queries: {time()-t0}s', src=0)
    log.info(f'acc for items in bf: {acc_list}', src=0)

    acc = 0
    for k in keys:
        if bf[hash((k,107))%B]==1:
            acc += 1
    log.info(f'acc for item not in bf: {acc}', src=0)


    one_share = protocol.share(src=0, secret=numpy.array(1, dtype=object), shape=())
    shared_bf = [protocol.share(src=0, secret=numpy.array(0, dtype=object), shape=()) for x in range(B)]
    #creating shared bloom filter
    t0=time()
    for i in items:
        for k in keys:
            bf_index = hash((k,i)) %B
            shared_bf[bf_index]=protocol.logical_or(one_share, shared_bf[bf_index])
    log.info(f'time to construct: {time()-t0}s', src=0)
    #querying the shared bloom filter for all the items it should contain by the leaky method
    t0=time()
    shared_acc_list = [protocol.share(src=0, secret=numpy.array(0, dtype=object), shape=()) for x in range(n)]
    for i,e in enumerate(items):
        for k in keys:
            bf_index = hash((k,e)) %B
            shared_acc_list[i] = protocol.add(shared_acc_list[i], shared_bf[bf_index])
    revealed_acc_list_sums = [int(protocol.reveal(shared_acc_list[i])) for i in range(n)]
    hundred_time = time()-t0
    log.info(f'time to make 100 leaky queries: {hundred_time}s, average time per leaky query: {hundred_time/100}s', src=0)
    log.info(f'acc for items in bf: {revealed_acc_list_sums}', src=0)

    t0=time()
    shared_acc = protocol.share(src=0, secret=numpy.array(0, dtype=object), shape=())
    for k in keys:
        bf_index = hash((k,107)) %B
        shared_acc = protocol.add(shared_acc, shared_bf[bf_index])
    revealed_acc = int(protocol.reveal(shared_acc))
    log.info(f'time to make 1 leaky query: {time()-t0}s', src=0)
    log.info(f'shared acc for item not in bf leaky query: {revealed_acc}', src=0)
    #querying the shared bloom filter for all the items it should contain by the leaky method
    t0=time()
    shared_prod_list = [protocol.share(src=0, secret=numpy.array(1, dtype=object), shape=()) for x in range(n)]
    for i,e in enumerate(items):
        for k in keys:
            bf_index = hash((k,e)) %B
            shared_prod_list[i] = protocol.untruncated_multiply(shared_prod_list[i], shared_bf[bf_index])
    revealed_prod_list = [int(protocol.reveal(shared_prod_list[i])) for i in range(n)]
    hundred_time = time()-t0
    log.info(f'time to make 100 less leaky queries: {hundred_time}s, average time per less leaky query: {hundred_time/100}s', src=0)
    log.info(f'acc for items in bf: {revealed_prod_list}', src=0)
    t0=time()
    shared_acc = protocol.share(src=0, secret=numpy.array(1, dtype=object), shape=())
    for k in keys:
        bf_index = hash((k,107)) %B
        shared_acc = protocol.untruncated_multiply(shared_acc, shared_bf[bf_index])
    revealed_acc = int(protocol.reveal(shared_acc))
    log.info(f'time to make 1 less leaky query: {time()-t0}s', src=0)
    log.info(f'shared acc for item not in bf less leaky method: {revealed_acc}', src=0)

cicada.communicator.SocketCommunicator.run(world_size=3, fn=main)




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import test

from behave import *
import numpy

import cicada.encoding


@given(u'a Bits encoding')
def step_impl(context):
    if "encodings" not in context:
        context.encodings = []
    context.encodings.append(cicada.encoding.Bits())


@given(u'a Boolean encoding')
def step_impl(context):
    if "encodings" not in context:
        context.encodings = []
    context.encodings.append(cicada.encoding.Boolean())


@given(u'a Identity encoding')
def step_impl(context):
    if "encodings" not in context:
        context.encodings = []
    context.encodings.append(cicada.encoding.Identity())


@given(u'a default FixedPoint encoding')
def step_impl(context):
    if "encodings" not in context:
        context.encodings = []
    context.encodings.append(cicada.encoding.FixedPoint())


@given(u'a {bits} bit FixedPoint encoding')
def step_impl(context, bits):
    bits = eval(bits)
    if "encodings" not in context:
        context.encodings = []
    context.encodings.append(cicada.encoding.FixedPoint(precision=bits))


@when(u'{value} is encoded and decoded')
def step_impl(context, value):
    value = eval(value)
    if value is not None:
        value = numpy.array(value)

    encoding = context.encodings[-1]
    field = context.fields[-1]

    encoded = encoding.encode(value, field)
    decoded = encoding.decode(encoded, field)
    context.decoded = decoded


@then(u'the encodings should compare equal')
def step_impl(context):
    lhs, rhs = context.encodings
    test.assert_true(lhs == rhs)


@then(u'the encodings should compare unequal')
def step_impl(context):
    lhs, rhs = context.encodings
    test.assert_true(lhs != rhs)


@then(u'the decoded value should match None')
def step_impl(context):
    test.assert_equal(context.decoded, None)


@then(u'the decoded value should match {value}')
def step_impl(context, value):
    value = numpy.array(eval(value))
    numpy.testing.assert_array_equal(context.decoded, value)


@then(u'the decoded value should be an array or None')
def step_impl(context):
    test.assert_is_instance(context.decoded, (numpy.ndarray, type(None)))


#def assert_is_fixed_field_representation(array):
#    test.assert_is_instance(array, numpy.ndarray)
#    test.assert_equal(array.dtype, object)
#    for value in array.flat:
#        test.assert_is_instance(value, int)
#
#
#@when(u'{x} is encoded the shape of the encoded array should be {shape}')
#def step_impl(context, x, shape):
#    x = numpy.array(eval(x))
#    shape = eval(shape)
#
#    encoder = context.encoders[-1]
#    encoded = encoder.encode(x)
#    assert_is_fixed_field_representation(encoded)
#    test.assert_equal(encoded.shape, shape)
#
#
#@when(u'{x} is encoded the shape of the decoded array should be {shape}')
#def step_impl(context, x, shape):
#    x = numpy.array(eval(x))
#    shape = eval(shape)
#
#    encoder = context.encoders[-1]
#    encoded = encoder.encode(x)
#    assert_is_fixed_field_representation(encoded)
#    decoded = encoder.decode(encoded)
#    test.assert_equal(decoded.shape, shape)
#
#@when(u'None is encoded and decoded the result should be None')
#def step_impl(context):
#    encoder = context.encoders[-1]
#    encoded = encoder.encode(None)
#    test.assert_is_none(encoded)
#    decoded = encoder.decode(encoded)
#    test.assert_is_none(decoded)
#
#
#@when(u'generating zeros with shape {shape} the result should match {result}')
#def step_impl(context, shape, result):
#    shape = eval(shape)
#    result = eval(result)
#
#    encoder = context.encoders[-1]
#    encoded = encoder.zeros(shape)
#    numpy.testing.assert_array_equal(encoded, result)
#
#
#@when(u'generating zeros like {other} the result should match {result}')
#def step_impl(context, other, result):
#    other = numpy.array(eval(other))
#    result = numpy.array(eval(result))
#
#    encoder = context.encoders[-1]
#    encoded = encoder.zeros_like(other)
#    numpy.testing.assert_array_equal(encoded, result)
#
#
#@when(u'{x} is encoded and decoded the result should match {y}')
#def step_impl(context, x, y):
#    x = numpy.array(eval(x), dtype=numpy.float64)
#    y = numpy.array(eval(y), dtype=numpy.float64)
#
#    encoder = context.encoders[-1]
#    encoded = encoder.encode(x)
#    assert_is_fixed_field_representation(encoded)
#    decoded = encoder.decode(encoded)
#
#    numpy.testing.assert_almost_equal(decoded, y, decimal=4)
#
#
#@when(u'matrix {A} and vector {x} are encoded and multiplied without truncation, the decoded result should match {y}')
#def step_impl(context, A, x, y):
#    encoder = context.encoders[-1]
#    A = encoder.encode(numpy.array(eval(A)))
#    x = encoder.encode(numpy.array(eval(x)))
#    y = numpy.array(eval(y))
#
#    result = encoder.untruncated_matvec(A,x)
#    assert_is_fixed_field_representation(result)
#    decoded = encoder.decode(result)
#    if not numpy.allclose(decoded, y, rtol=0, atol=0.001):
#       raise ValueError("result mismatch")
#
#
#@when(u'{b} is subtracted from {a} the result should match {c}')
#def step_impl(context, a, b, c):
#    encoder = context.encoders[-1]
#    a = encoder.encode(numpy.array(eval(a)))
#    b = encoder.encode(numpy.array(eval(b)))
#    c = numpy.array(eval(c))
#
#    result = encoder.subtract(a, b)
#    assert_is_fixed_field_representation(result)
#    decoded = encoder.decode(result)
#    numpy.testing.assert_almost_equal(decoded, c, decimal=4)
#
#
#
#@when(u'{a} is negated the result should match {b}')
#def step_impl(context, a, b):
#    encoder = context.encoders[-1]
#    a = encoder.encode(numpy.array(eval(a)))
#    b = numpy.array(eval(b))
#
#    result = encoder.negative(a)
#    assert_is_fixed_field_representation(result)
#    decoded = encoder.decode(result)
#    numpy.testing.assert_almost_equal(decoded, b, decimal=4)
#
#
#@when(u'{a} is summed the result should match {b}')
#def step_impl(context, a, b):
#    encoder = context.encoders[-1]
#    a = encoder.encode(numpy.array(eval(a)))
#    b = numpy.array(eval(b))
#
#    result = encoder.sum(a)
#    assert_is_fixed_field_representation(result)
#    decoded = encoder.decode(result)
#    numpy.testing.assert_almost_equal(decoded, b, decimal=4)
#



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest.mock

from behave import *

import cicada.interactive
from cicada.communicator import SocketCommunicator


@when(u'player {} receives secret input {}')
def step_impl(context, player, text):
    player = eval(player)
    text = eval(text)

    def operation(communicator, player, text):
        cicada.interactive.input = unittest.mock.MagicMock(return_value=text)
        return cicada.interactive.secret_input(communicator=communicator, src=player)

    context.results = SocketCommunicator.run(world_size=context.players, fn=operation, args=(player, text))


